# -*- coding: utf-8 -*-
#
# Copyright (c) 2017 Muges
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
The :mod:`erika.frontend.widgets.webview` provides a webview that is styled
using the current gtk theme.
"""

import os
from tempfile import NamedTemporaryFile
from string import Template
import webbrowser

from gi.repository import Gst
from gi.repository import Gtk
from gi.repository import Pango
from gi.repository import WebKit


class WebView(WebKit.WebView):
    """A webview styled using the current gtk theme."""
    def __init__(self):
        super().__init__()

        # Make the webview transparent (to use default background color)
        screen = self.get_screen()
        rgba = screen.get_rgba_visual()
        self.set_visual(rgba)

        self.set_transparent(True)

        # Disable scripts, plugins and context menu
        settings = self.get_settings()
        settings.set_property('enable-java-applet', False)
        settings.set_property('enable-plugins', False)
        settings.set_property('enable-scripts', False)
        settings.set_property('enable-default-context-menu', False)

        self.first_draw = True
        self.stylesheet_filename = None

        self.connect('navigation-policy-decision-requested',
                     self._on_link_clicked)
        self.connect('destroy', self._on_destroy)

    def do_draw(self, cr):
        if self.first_draw:
            context = self.get_style_context()
            settings = self.get_settings()

            # Set default font
            font = context.get_font(Gtk.StateFlags.NORMAL)
            settings.set_property('default-font-size',
                                  font.get_size() // Pango.SCALE)
            settings.set_property('default-font-family', font.get_family())

            # Generate stylesheet with default colors
            normal_color = context.get_color(Gtk.StateFlags.NORMAL).to_string()
            link_color = context.get_color(Gtk.StateFlags.LINK).to_string()

            stylesheet = Template("""
            body {
                color: $normal_color;
            }
            a {
                color: $link_color;
            }
            """).substitute(normal_color=normal_color, link_color=link_color)

            # Store the stylesheet in a temporary file
            # (the user-stylesheet-uri property does not support data uris)
            with NamedTemporaryFile(suffix='.css', delete=False) as fileobj:
                fileobj.write(stylesheet.encode('utf8'))
                self.stylesheet_filename = fileobj.name

            stylesheet_uri = Gst.filename_to_uri(self.stylesheet_filename)
            settings.set_property('user-stylesheet-uri', stylesheet_uri)

            self.first_draw = False

        return WebKit.WebView.do_draw(self, cr)

    def _on_link_clicked(self, view, frame, request, action, decision):
        """Called when a link is clicked"""
        # pylint: disable=too-many-arguments,no-self-use
        webbrowser.open(request.get_uri())
        decision.ignore()
        return True

    def _on_destroy(self, _):
        """Called when the widget is destroyed"""
        # Remove the temporary stylesheet
        os.remove(self.stylesheet_filename)



# -*- coding: utf-8 -*-
#
# Copyright (c) 2016 Muges
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
The :mod:`erika.config` module defines constants used for the configuration or
erika.
"""

from os.path import expanduser, join
import platform
from gi.repository import GLib

from .__version__ import __appname__


# Directories
HOME = expanduser("~")
DEFAULT_CONFIG_DIRECTORY = join(GLib.get_user_config_dir(), __appname__.lower())

CONFIG_DEFAULTS = {
    "application.version": (0,),

    "library.root": join(HOME, "Podcasts"),
    "library.podcast_directory_template": "{podcast.title}",
    "library.episode_file_template":
        "{episode.pubdate:%Y.%m.%d} - {episode.title}",
    "library.synchronize_interval": 60,

    "player.smart_mark_seconds": 30,

    "downloads.workers": 2,

    "gpodder.synchronize": False,
    "gpodder.hostname": "gpodder.net",
    "gpodder.username": "",
    "gpodder.password": "",
    "gpodder.deviceid": "{}-{}".format(__appname__.lower(), platform.node()),
    "gpodder.devicename": "{} on {}".format(__appname__, platform.node()),
    "gpodder.devicename_changed": False,
    "gpodder.last_subscription_sync": 0,
    "gpodder.last_episodes_sync": 0,

    "network.connection_check_hostname": "github.com",
}



# -*- coding: utf-8 -*-
#
# Copyright (c) 2016 Muges
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
A model used to store episodes in the library.
"""

import logging
import os.path
import requests
from peewee import (BooleanField, DateTimeField, IntegerField, DeferredForeignKey,
                    TextField, Proxy, DoesNotExist)
from playhouse.hybrid import hybrid_property

from erika import tags
from erika.library.fields import Image
from erika.util import format_duration, sanitize_filename
from .database import BaseModel
from .config import Config

Podcast = Proxy()  # pylint: disable=invalid-name


class Episode(BaseModel):
    """
    A model used to store episodes in the library.

    Attributes
    ----------
    podcast : Podcast
        The podcast the episode belongs to.
    guid : str, optional
        A unique identifier for the episode.
    pubdate : datetime, optional
        The date of publication of the episode.
    title : str, optional
        The title of the episode.
    duration : int, optional
        The duration of the episode in seconds.
    image_url : str, optional
        A link to the image of the episode.
    subtitle : str, optional
        The subtitle of the episode in plaintext.
    summary : str, optional
        A description of the episode in html.
    link : str, optional
        Website link.
    track_number : int
        The track number of the episode. As long as the publication dates are
        consistent with the publication order this should be accurate.
    file_url : str, optional
        Url of the episode file.
    file_size : int, optional
        Size of the episode file in bytes.
    mimetype : str, optional
        Mimetype of the episode file.
    local_path : str, optional
        Path of the downloaded file.
    new : bool
        True if the episode is new.
    played : bool
        True if the episode has been played.
    progress : int
        Number of seconds of the episode that have been played (used to be able
        to resume the playback after quitting the application).
    """

    podcast = DeferredForeignKey('Podcast', related_name='episodes',
                                 on_delete='CASCADE')

    guid = TextField(null=True)
    pubdate = DateTimeField(null=True)
    title = TextField(null=True)
    duration = IntegerField(null=True)
    image_url = TextField(null=True)
    subtitle = TextField(null=True)
    summary = TextField(null=True)
    link = TextField(null=True)

    track_number = IntegerField()

    file_url = TextField(null=True, index=True)
    file_size = IntegerField(null=True)
    mimetype = TextField(null=True)
    local_path = TextField(null=True)

    new = BooleanField(default=True)
    played = BooleanField(default=False)
    progress = IntegerField(default=0)

    class Meta:  # pylint: disable=too-few-public-methods, missing-docstring
        indexes = (
            (('podcast', 'guid'), True),
            (('podcast', 'title', 'pubdate'), True),
        )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._image = None
        self._tried_image_tags = False

    @property
    def absolute_local_path(self):
        """str: the absolute path of the local episode's file."""
        if self.local_path:
            library_root = Config.get_value("library.root")
            return os.path.abspath(os.path.join(library_root, self.local_path))
        else:
            return None

    @absolute_local_path.setter
    def absolute_local_path(self, path):
        if path:
            library_root = Config.get_value("library.root")
            self.local_path = os.path.relpath(path, library_root)
        else:
            self.local_path = None

    @property
    def display_duration(self):
        """str: the episode's duration as a formatted string."""
        if not self.duration:
            return ""

        return format_duration(self.duration)

    @property
    def image(self):
        """:class:`~erika.library.fields.Image`: the episode's image."""
        if self._image:
            return self._image

        # Try to get the image from the audio file
        if self.local_path and not self._tried_image_tags:
            self._tried_image_tags = True

            audio_tags = tags.get_tags(self.absolute_local_path)
            if audio_tags["image"]:
                self._image = Image(audio_tags["image"])
                return self._image

        return self.podcast.image

    @property
    def image_downloaded(self):
        """bool: true if the episode's image has been downloaded."""
        return self._image

    @hybrid_property
    def downloaded(self):
        """bool: true if the episode has been downloaded."""
        return self.local_path is not None

    @downloaded.expression
    def downloaded(self):
        """bool: true if the episode has been downloaded."""
        return self.local_path.is_null(False)

    @classmethod
    def get_matching(cls, path):
        """Get the episode matching a file and move the file to the right
        location.

        Parameters
        ----------
        path : str
            The path of the file.

        Returns
        -------
        :class:`Episode`, optional
            The episode, or None if there was no match.
        """
        logger = logging.getLogger(".".join((__name__, cls.__name__)))
        logger.debug("Searching for an episode matching the file '%s'", path)

        # Read the file's audio metadata
        audiotags = tags.get_tags(path)
        if not audiotags:
            return None

        podcast_title = audiotags['podcast']

        # Try to find the podcast
        try:
            podcast = Podcast.get(Podcast.title == podcast_title)
        except DoesNotExist:
            logger.warning("No podcast with title %s.", podcast_title)
            return None

        # Find the episode matching the metadata stricty
        episode = podcast.get_episode_from_tags_strict(audiotags)
        if episode:
            return episode

        # Try to find an episode matching the metadata loosely
        filename = os.path.splitext(os.path.basename(path))[0]
        episode = podcast.get_episode_from_tags_loose(audiotags, filename)
        if episode:
            episode.import_file(path)
            return episode

        logger.warning("No match found.")

    @staticmethod
    def get_counts():
        """Return a tuple containing some statistics about all the episodes.

        Returns
        -------
        int
            The number of new episodes.
        int
            The number of played episodes.
        int
            The total number of episodes.
        """
        # pylint: disable=singleton-comparison
        new = Episode.select().where(Episode.new == True).count()
        played = Episode.select().where(Episode.played == True).count()
        total = Episode.select().count()
        return new, played, total

    def import_file(self, path):
        """Import the episode's audio file in the library.

        Moves the file and sets its tag."""
        extension = os.path.splitext(path)[1]
        new_path = self.get_default_filename(extension)

        os.rename(path, new_path)
        tags.set_tags(new_path, self)

        return new_path

    def get_default_filename(self, extension):
        """Return the default filename for the audio file of an episode given
        its extension.

        Parameters
        ----------
        extension : str

        Returns
        -------
        str
            The default filename for the episode.
        """
        dirname = self.podcast.get_directory()
        template = Config.get_value("library.episode_file_template")
        filename = os.path.join(
            dirname,
            sanitize_filename(template.format(episode=self))
        )
        return filename + extension

    def mark_as_played(self):
        """Mark the episode as played."""
        self.played = True
        self.new = False

    def mark_as_unplayed(self):
        """Mark the episode as unplayed."""
        self.played = False

    def get_image(self):
        """Get the episode's image.

        This method downloads the episode's image if ``image_url`` is not None
        (and it has not been dowloaded yet), and falls back to the podcast's
        image if it is None or if an error occured.

        Returns
        -------
        :class:`~erika.library.fields.Image`
            The image.
        """
        # Return the episode's image if it has already been downloaded
        if self._image:
            return self._image

        if self.image_url:
            # Download image
            try:
                response = requests.get(self.image_url)
                response.raise_for_status()
            except requests.exceptions.RequestException:
                self.logger.exception("Unable to download image.")
            else:
                self._image = Image(response.content)
                return self._image

        # Fallback to the podcast's image
        return self.podcast.image



# -*- coding: utf-8 -*-
#
# Copyright (c) 2016 Muges
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Widget allowing to easily display loading messages in a status bar
"""

# pylint: disable=arguments-differ

from collections import OrderedDict

from gi.repository import Gtk


class StatusBox(Gtk.HBox):
    """Widget allowing to easily display loading messages in a status bar

    The add method adds a new message to the StatusBox. The messages are
    displayed in the order they where added. As long as at least one message
    is being displayed, a spinner is displayed at the start (usually left) of
    the StatusBox.

    Each message can be removed by calling the remove method with the
    corresponding message id, which is returned by the add method.
    """
    def __init__(self):
        Gtk.HBox.__init__(self)
        self.set_spacing(5)

        self.next_id = 1
        self.messages = OrderedDict()
        self.separators = []

        self.spinner = Gtk.Spinner()
        self.pack_start(self.spinner, False, False, 0)

    def _get_message_index(self, message_id):
        """Get the position of a message given its id"""
        for index, key in enumerate(self.messages.keys()):
            if key == message_id:
                return index

    def get_next_message_id(self):
        """Return the id of the next message"""
        message_id = self.next_id
        self.next_id += 1
        return message_id

    def add(self, message, message_id):
        """Add a new message

        Parameters
        ----------
        message : str
            The message that will be displayed.
        message_id : int
            The id of the message (this should be the value returned by
            self.get_next_message_id())
        """
        # Add a separator between this message and the previous one if needed
        if self.messages:
            separator = Gtk.Separator.new(Gtk.Orientation.VERTICAL)
            separator.show()
            self.pack_start(separator, False, False, 0)
            self.separators.append(separator)

        # Add a label displaying the message
        label = Gtk.Label(message)
        label.show()
        self.pack_start(label, False, False, 0)
        self.messages[message_id] = label

        # Start the spinner
        self.spinner.start()

    def edit(self, message_id, message):
        """Edit a message

        Parameters
        ----------
        message_id : int
            The id of the message.
        """
        # Edit the label
        label = self.messages[message_id]
        label.set_text(message)

    def remove(self, message_id):
        """Remove a message

        Parameters
        ----------
        message_id : int
            The id of the message.
        """
        index = self._get_message_index(message_id)
        if index is None:
            # The message does not exist
            return

        # Remove the label
        label = self.messages.pop(message_id)
        label.destroy()  # pylint: disable=no-member

        # Remove the separator
        try:
            separator = self.separators.pop(max(0, index - 1))
        except IndexError:
            pass
        else:
            separator.destroy()

        # Stop the spinner if no message is being displayed
        if not self.messages:
            self.spinner.stop()



# -*- coding: utf-8 -*-
#
# Copyright (c) 2016 Muges
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
The :mod:`erika.parsers.rss` module implements a parser for RSS and Atom feeds.
"""

import logging
from time import mktime
from datetime import datetime
import feedparser
import requests

from erika.library import Episode
from erika.util import plaintext_to_html, html_to_plaintext

MIN_SUBTITLE_LENGTH = 25


def parse_feed(feed, podcast):
    """Parse the feed element."""
    try:
        image = feed.image.href
    except AttributeError:
        image = None

    if image != podcast.image_url:
        # Only update the podcast image if it changed, to avoid
        # setting the field as dirty
        podcast.image_url = image

    podcast.title = feed.get("title", None)
    podcast.author = feed.get("author", None)
    podcast.language = feed.get("language", None)
    podcast.subtitle = feed.get("subtitle", None)
    podcast.summary = feed.get("summary", None)
    podcast.link = feed.get("link", None)


def parse_links(links):
    """Parse an entry's links to find the audio file.

    Returns
    -------
    Optional[str]
        The url of the audio file
    Optional[int]
        The size of the audio file
    Optional[str]
        The mimetype of the audio file
    """
    logger = logging.getLogger(__name__)

    for link in links:
        # Check that the link is an enclosure
        if "rel" not in link or link.rel != "enclosure":
            continue

        # Check that the link is valid
        if "href" not in link:
            logger.warning("Skipping enclosure with no href attribute.")
            continue

        return link.href, link.get("length", None), link.type

    logger.warning("No enclosure found.")
    return None, None, None


def parse_duration(string):
    """Parse a duration of the form ``[hours:]minutes:seconds``."""
    if string == "":
        return None

    duration = string.split(":")[:3]

    try:
        duration = [int(i) for i in duration]
    except ValueError:
        logger = logging.getLogger(__name__)
        logger.warning("Invalid duration : \"%s\".", string)
        return None

    hours, minutes, seconds = [0] * (3 - len(duration)) + duration

    return hours * 3600 + minutes * 60 + seconds


def summary_to_subtitle(summary):
    """Convert a summary (in html) to a subtitle (in plaintext)."""
    summary_lines = [
        line.strip()
        for line in html_to_plaintext(summary).split("\n")
        if line.strip()
    ]

    # Get the first lines, until the total length is bigger than
    # MIN_SUBTITLE_LENGTH
    length = -1
    end = 0
    for line in summary_lines:
        length += 1 + len(line)
        end += 1
        if length > MIN_SUBTITLE_LENGTH:
            break

    return " ".join(summary_lines[:end])


def parse_entry(entry):
    """Parse an rss entry."""
    if "published_parsed" in entry:
        timestamp = mktime(entry.published_parsed)
        pubdate = datetime.fromtimestamp(timestamp)
    else:
        pubdate = None

    default_guid = "{} - {}".format(pubdate, entry.get("title", ""))

    try:
        image = entry.image.href
    except AttributeError:
        image = None

    file_url, file_size, mimetype = parse_links(entry.links)

    # Get summary
    try:
        if entry.summary_detail['type'] == "text/plain":
            summary = plaintext_to_html(entry["summary"])
        else:
            summary = entry["summary"]
    except AttributeError:
        summary = None

    # Create subtitle from summary (this usually gives better subtitles than
    # the ones in the rss feed)
    subtitle = summary_to_subtitle(summary)

    # Subtitle fallback
    if not subtitle:
        subtitle = entry.get("subtitle", None)

    # Summary fallback
    if not summary:
        summary = plaintext_to_html(subtitle)

    return Episode(
        guid=entry.get("id", default_guid),
        pubdate=pubdate,
        title=entry.get("title", None),
        duration=parse_duration(entry.get("itunes_duration", "")),
        image_url=image,
        subtitle=subtitle,
        summary=summary,
        link=entry.get("link", None),

        file_url=file_url,
        file_size=file_size,
        mimetype=mimetype
    )


def parse(podcast):
    """Parse an RSS or Atom feed.

    Parameters
    ----------
    podcast : :class:`Podcast`
        The podcast to parse.

    Returns
    -------
    list of :class:`Episode`
        A list of the podcast's episodes.
    """
    logger = logging.getLogger(__name__)
    logger.debug("Parsing %s.", podcast.url)

    response = requests.get(podcast.url, timeout=10)
    document = feedparser.parse(response.content)

    parse_feed(document.feed, podcast)
    episodes = [parse_entry(entry) for entry in document.entries]

    return episodes



# -*- coding : utf-8 -*-

import string
import pytest

from podcasts.library.models.database import slugify


@pytest.mark.parametrize("name, slug", [
    # Lowercase alphanumeric strings
    ("", ""),
    ("abcdef", "abcdef"),
    ("abc123", "abc123"),
    ("abc123", "abc123"),
    (string.ascii_lowercase, string.ascii_lowercase),
    (string.digits, string.digits),

    # Alphanumeric strings
    ("aAbBcCdD1", "aabbccdd1"),
    (string.ascii_uppercase, string.ascii_lowercase),

    # Non-alphanumeric strings
    ("abc-def", "abcdef"),
    ("Abc.D1f", "abcd1f"),
    ("Abc D1f", "abcd1f"),
    (string.punctuation, ""),
    (string.whitespace, ""),
])
def test_slugify(name, slug):
    """Test for slugify"""
    assert slugify(name) == slug



# -*- coding: utf-8 -*-
#
# Copyright (c) 2016 Muges
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
The :mod:`erika.parsers` module implements parsers for the sources of podcasts.

A parser is a submodule of the :mod:`erika.parsers` module. It should implement
a ``parse`` function, which, given a podcast, parses its source, set its
attributes, and returns a list of its episodes (see the
:mod:`erika.parsers.rss` module for an example).

Each :class:`Podcast` has an attribute ``parser``, which is the name of the
submodule that should be used.

For now the only available is the :mod:`erika.parsers.rss` module, for RSS and
Atom feeds, and support for different parsers has not yet been added to the
GUI.
"""

import importlib


def parse(podcast):
    """Parse a podcast source, edit the podcast's attribute, and return a list
    of the podcast's episode.

    Parameters
    ----------
    podcast : :class:`Podcast`
        The podcast to parse.

    Returns
    -------
    list of :class:`Episode`
        A list of the podcast's episodes.
    """
    if "." in podcast.parser:
        raise ValueError(
            "{} is not a valid parser name.".format(podcast.parser))

    try:
        parser = importlib.import_module(".".join((__name__, podcast.parser)))
    except ModuleNotFoundError:
        raise ValueError(
            "{} is not a valid parser name.".format(podcast.parser))

    episodes = parser.parse(podcast)

    for episode in episodes:
        episode.podcast = podcast

    return sorted(episodes, key=lambda e: e.pubdate)



# -*- coding: utf-8 -*-
#
# Copyright (c) 2016 Muges
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
A button used to represent sorting options
"""

from gi.repository import Gtk


class SortButton(Gtk.Button):
    """A button used to represent sorting options

    The button has two states : ascending or descending.
    """
    def __init__(self, sort_by):
        Gtk.Button.__init__(self)
        self.connect("clicked", SortButton._on_clicked)

        self.sort_by = sort_by
        self.descending = True

        self.ascending_image = Gtk.Image.new_from_icon_name(
            "view-sort-ascending-symbolic", Gtk.IconSize.SMALL_TOOLBAR)
        self.descending_image = Gtk.Image.new_from_icon_name(
            "view-sort-descending-symbolic", Gtk.IconSize.SMALL_TOOLBAR)

        self.set_descending(True)

    def set_descending(self, descending):
        """Change the state of the button

        Parameters
        ----------
        descending : bool
            True if the sorting order whould be descending.
        """
        self.descending = descending

        if descending:
            self.set_image(self.descending_image)
            self.set_tooltip_text("Sort by ascending {}".format(self.sort_by))
        else:
            self.set_image(self.ascending_image)
            self.set_tooltip_text("Sort by descending {}".format(self.sort_by))

    def get_descending(self):
        """Return the state of the button

        Returns
        -------
        bool
            True if the sorting order whould be descending.
        """
        return self.descending

    def _on_clicked(self):
        """Called when the button is clicked

        Toggle the state of the button.
        """
        self.set_descending(not self.get_descending())



# -*- coding: utf-8 -*-
#
# Copyright (c) 2016 Muges
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
The :mod:`erika.library.opml` module provides functions to import and export
opml files.
"""

from email.utils import formatdate
from lxml import etree

from .models import Podcast


def import_opml(filename):
    """
    Import podcasts subscriptions from an opml file.

    Parameters
    ----------
    filename : str
        The opml file.
    """
    document = etree.parse(filename)

    for outline in document.xpath("/opml/body/outline"):
        Podcast.new(outline.get("type", "rss"), outline.get("xmlUrl"))


def export_opml(filename):
    """
    Export the podcasts subscriptions to an opml file.

    Parameters
    ----------
    filename : str
        The opml file.
    """
    root = etree.Element('opml')
    document = etree.ElementTree(root)

    # Head
    head = etree.SubElement(root, 'head')

    title = etree.SubElement(head, 'title')
    title.text = "Podcasts subscriptions"

    date = etree.SubElement(head, 'dateCreated')
    date.text = formatdate()

    # Body
    body = etree.SubElement(root, 'body')

    for podcast in Podcast.select():
        etree.SubElement(body, 'outline',
                         text=podcast.title,
                         title=podcast.title,
                         type=podcast.parser,
                         xmlUrl=podcast.url,
                         htmlUrl=podcast.link)

    document.write(filename, xml_declaration=True, encoding='utf-8',
                   pretty_print=True)



# -*- coding: utf-8 -*-
#
# Copyright (c) 2016 Muges
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Module handling the audio playback
"""

import logging
import os

from gi.repository import GObject
from gi.repository import Gst
from gi.repository import GLib

from erika.library.models import Config, EpisodeAction

# For some reason this is not defined in the python gstreamer bindings.
GST_PLAY_FLAG_DOWNLOAD = 1 << 7


class Player(GObject.Object):
    """
    Object handling the audio playback

    Signals
    -------
    episode-updated(Episode)
        Emitted at the end of the playback of an episode.
    progress-changed(position, duration)
        Emitted every 200ms during playback.
    state-changed(episode, state)
        Emitted when the player state has changed.
    """

    STOPPED = 0
    PLAYING = 1
    PAUSED = 2
    BUFFERING = 3

    __gsignals__ = {
        'episode-updated':
            (GObject.SIGNAL_RUN_FIRST, None, (GObject.TYPE_PYOBJECT,)),
        'progress-changed':
            (GObject.SIGNAL_RUN_FIRST, None, (GObject.TYPE_INT64,
                                              GObject.TYPE_INT64)),
        'state-changed':
            (GObject.SIGNAL_RUN_FIRST, None, (GObject.TYPE_PYOBJECT,
                                              GObject.TYPE_INT,)),
    }

    def __init__(self):
        GObject.Object.__init__(self)

        self.logger = logging.getLogger(
            "{}.{}".format(__name__, self.__class__.__name__))

        self.episode = None
        self.state = Player.STOPPED
        self._started = 0
        self._duration = 0

        self.progress_timer = None

        # Create player
        self.player = Gst.ElementFactory.make("playbin")
        flags = self.player.get_property("flags")
        flags |= GST_PLAY_FLAG_DOWNLOAD
        self.player.set_property("flags", flags)

        bus = self.player.get_bus()
        bus.add_signal_watch()
        bus.connect("message", self._on_message)

    def play_episode(self, episode):
        """Start the playback of an episode

        Parameters
        ----------
        episode : Episode
        """
        self.stop()
        self.episode = episode
        self._started = self.episode.progress

        self.set_state(Player.BUFFERING)

        if episode.local_path and os.path.isfile(episode.absolute_local_path):
            uri = Gst.filename_to_uri(episode.absolute_local_path)
        else:
            uri = episode.file_url

        self.player.set_property("uri", uri)
        self.player.set_state(Gst.State.PLAYING)

        self.progress_timer = GObject.timeout_add(200, self._emit_progress)

    def stop(self):
        """Stop the playback of an episode"""
        if self.progress_timer:
            GLib.source_remove(self.progress_timer)
            self.progress_timer = None

        if not self.episode:
            return

        self.set_state(Player.STOPPED)

        smart_mark_seconds = Config.get_value("player.smart_mark_seconds")

        position = self.get_position() // Gst.SECOND
        duration = self.get_duration() // Gst.SECOND
        if position + smart_mark_seconds >= duration:
            # Mark as read and unset progress
            self.episode.progress = 0
            self.episode.mark_as_played()
            position = duration
        else:
            # Save progress
            self.episode.progress = position
        self.episode.save()

        action = EpisodeAction(episode=self.episode, action="play",
                               started=self._started, position=position,
                               total=duration)
        action.save()

        self.emit("episode-updated", self.episode)

        self.player.set_state(Gst.State.NULL)
        self.episode = None
        self._duration = 0

    def get_duration(self):
        """Return the duration of the current episode, or 0 if it is unknown"""
        if not self._duration:
            success, duration = self.player.query_duration(Gst.Format.TIME)
            if success:
                self._duration = duration

        return self._duration

    def get_position(self):
        """Return the current playback position, or 0 if it is unknown"""
        success, position = self.player.query_position(Gst.Format.TIME)
        if success:
            return position

        return 0

    def get_seconds_duration(self):
        """Return the duration of the current episode in seconds, or 0 if it
        is unknown"""
        return self.get_duration() // Gst.SECOND

    def get_seconds_position(self):
        """Return the current playback position in seconds, or 0 if it is
        unknown"""
        return self.get_position() // Gst.SECOND

    def percent_to_time(self, percent):
        """Convert a time in format Gst.Format.PERCENT to Gst.Format.TIME

        Parameters
        ----------
        percent : int
        """
        duration = self.get_duration()
        return percent * duration / 1000000

    def get_buffered(self):
        """Return the percentage of the episode that has been buffered"""
        position = self.player.query_position(Gst.Format.PERCENT)[1] or 0

        query = Gst.Query.new_buffering(Gst.Format.PERCENT)
        if self.player.query(query):
            stop = query.parse_buffering_range()[2]
            if stop >= position:
                return self.percent_to_time(stop)

            for range_index in range(0, query.get_n_buffering_ranges()):
                stop = query.parse_nth_buffering_range(range_index)[2]
                if stop >= position:
                    return self.percent_to_time(stop)

        return self.get_position()

    def seek(self, position):
        """Send a seek event to the player

        Parameters
        ----------
        position : int
            Position in nanoseconds
        """
        if position < 0:
            position = 0

        self.player.seek_simple(
            Gst.Format.TIME,
            Gst.SeekFlags.FLUSH | Gst.SeekFlags.KEY_UNIT,
            position)

    def seek_relative(self, offset):
        """Send a seek event relative to the current position

        Parameters
        ----------
        offset : int
            Offset in nanoseconds
        """
        position = self.player.query_position(Gst.Format.TIME)[1] or 0
        position += offset
        self.seek(position)

    def _on_message(self, bus, message):
        """Handle gstreamer bus messages"""
        if message.type == Gst.MessageType.EOS:
            # End of stream
            self.stop()
        elif message.type == Gst.MessageType.ERROR:
            # The stream ended because of an error
            self.stop()

            err, _ = message.parse_error()
            self.logger.error(err.message)
            self.logger.error("Debugging info: %s", err.message)
        elif message.type == Gst.MessageType.DURATION_CHANGED:
            # The duration of the stream changed
            success, duration = self.player.query_duration(Gst.Format.TIME)
            if success:
                self._duration = duration
        elif message.type == Gst.MessageType.STATE_CHANGED:
            oldstate, newstate, _ = message.parse_state_changed()
            if oldstate == newstate:
                return

            if message.src != self.player:
                return

            if all((oldstate == Gst.State.READY,
                    newstate == Gst.State.PAUSED,
                    self.episode.progress > 0)):
                self.seek(self.episode.progress * Gst.SECOND)

            if newstate == Gst.State.READY:
                self.set_state(Player.BUFFERING)
            if newstate == Gst.State.PLAYING:
                self.set_state(Player.PLAYING)
            elif newstate == Gst.State.PAUSED:
                self.set_state(Player.PAUSED)

    def _emit_progress(self):
        """Emit a progress-changed signal. Called every 200ms during
        playback"""
        position = self.get_position()
        duration = self.get_duration()
        self.emit('progress-changed', position, duration)

        return True

    def set_state(self, state):
        """Set the current state of the player"""
        self.state = state
        self.emit('state-changed', self.episode, state)

    def play(self):
        """Start the playback"""
        self.player.set_state(Gst.State.PLAYING)

    def pause(self):
        """Pause the playback"""
        self.player.set_state(Gst.State.PAUSED)

    def toggle(self):
        """Toggle between play and pause"""
        if self.state == Player.PLAYING or self.state == Player.BUFFERING:
            self.pause()
        else:
            self.play()



# -*- coding: utf-8 -*-
#
# Copyright (c) 2016 Muges
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Table containing the episode actions
"""

from datetime import datetime
from peewee import DateTimeField, IntegerField, ForeignKeyField, TextField

from mygpoclient.api import EpisodeAction as GpoEpisodeAction
from mygpoclient.util import datetime_to_iso8601

from .database import BaseModel
from .episode import Episode
from .config import Config


class EpisodeAction(BaseModel):
    """A model used to store episode actions in the library.

    The episode actions are used for the gpodder synchronization.

    Attributes
    ----------
    episode : :class:`Episode`
    action : str
        The type of the action (download, delete, play or new).
    time : datetime, optional
        The UTC time when the action took place as a datetime.
    started : int, optional
        The position at which the playback started (only valid if action is
        'play').
    position : int, optional
        The position at which the playback was stopped (only valid if action is
        'play').
    total : int, optional
        The duration of the episode in seconds (only valid if action is
        'play').
    """
    episode = ForeignKeyField(Episode, on_delete='CASCADE')
    action = TextField()
    time = DateTimeField(default=datetime.utcnow)

    started = IntegerField(null=True)
    position = IntegerField(null=True)
    total = IntegerField(null=True)

    @property
    def podcast_url(self):
        """str: the url of the episode's podcast."""
        return self.episode.podcast.url

    @property
    def episode_url(self):
        """str: the url of the episode."""
        return self.episode.file_url

    @property
    def timestamp(self):
        """str: the UTC time when the action took place as an ISO8601
        timestamp."""
        return datetime_to_iso8601(self.time)

    def for_gpodder(self):
        """Convert the EpisodeAction into a
        :class:`mygpoclient.api.EpisodeAction`"""
        device = Config.get_value("gpodder.deviceid")

        if self.action == "play":
            return GpoEpisodeAction(self.podcast_url, self.episode_url,
                                    self.action, device, self.timestamp,
                                    self.started, self.position, self.total)
        else:
            return GpoEpisodeAction(self.podcast_url, self.episode_url,
                                    self.action, device, self.timestamp)

    def __str__(self):
        attrs = ["podcast_url", "episode_url", "action", "timestamp",
                 "started", "position", "total"]
        formatted_attrs = ["{}={}".format(attr, repr(getattr(self, attr)))
                           for attr in attrs]
        return "EpisodeAction({})".format(", ".join(formatted_attrs))



# -*- coding: utf-8 -*-
#
# Copyright (c) 2016 Muges
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
The :mod:`erika.tags.mp3` module provides functions to set and get tags of mp3
files.
"""

from datetime import datetime
from mutagen.id3 import (  # pylint: disable=no-name-in-module
    TPE1, TALB, TIT2, TSOT, TDRC, ID3TimeStamp, COMM, WOAF, WFED, WORS, TXXX,
    APIC, TRCK, TCON, Encoding, PictureType
)

ENCODING = Encoding.UTF8


def set_tags(audio, episode):  # pylint: disable=too-many-branches
    """Set the tags of the mp3 audio file of an episode.

    Parameters
    ----------
    filename : str
        The audio file.
    episode : :class:`Episode`
        The corresponding episode.
    """
    if not audio.tags:
        audio.add_tags()

    # Author and titles
    if episode.podcast.author:
        audio.tags["TPE1"] = TPE1(ENCODING, episode.podcast.author)
    if episode.podcast.title:
        audio.tags["TALB"] = TALB(ENCODING, episode.podcast.title)
    if episode.title:
        audio.tags["TIT2"] = TIT2(ENCODING, episode.title)
        audio.tags["TSOT"] = TSOT(ENCODING, episode.title)

    # Publication date
    if episode.pubdate:
        timestamp = ID3TimeStamp(episode.pubdate.isoformat(' '))
        audio.tags["TDRC"] = TDRC(ENCODING, [timestamp])

    # Description
    if "COMM" not in audio.tags and episode.summary:
        audio.tags["COMM"] = COMM(ENCODING, desc="", text=episode.summary)

    # Links
    if episode.link:
        audio.tags["WOAF"] = WOAF(episode.link)
    if episode.podcast.url:
        audio.tags["WFED"] = WFED(episode.podcast.url)
    if episode.podcast.link:
        audio.tags["WORS"] = WORS(episode.podcast.link)
    if episode.guid:
        audio.tags["TXXX"] = TXXX(ENCODING, desc="", text=episode.guid)

    # Track number
    audio.tags["TRCK"] = TRCK(ENCODING, str(episode.track_number))

    # Image
    if "APIC:" not in audio.tags:
        image = episode.get_image()

        if image:
            data = image.get_data()
            audio.tags["APIC"] = APIC(ENCODING, desc="", mime='image/png',
                                      type=PictureType.COVER_FRONT, data=data)

    # Genre
    audio.tags["TCON"] = TCON(ENCODING, text="Podcast")

    audio.save()


def get_tag(audio, tag):
    """Get an audio tag, or return None if it is not available."""
    try:
        return audio.tags[tag].text[0]
    except (KeyError, IndexError, TypeError):
        return None


def get_tags(audio):
    """Return the tags of an audio file."""
    date = get_tag(audio, "TDRC")
    try:
        date = datetime.strptime(date.text, "%Y-%m-%d %H:%M:%S")
    except (ValueError, AttributeError):
        date = None

    try:
        image = audio.tags["APIC:"].data
    except KeyError:
        image = None

    return {
        "podcast": get_tag(audio, "TALB"),
        "title": get_tag(audio, "TIT2"),
        "guid": get_tag(audio, "TXXX"),
        "pubdate": date,
        "image": image
    }



# -*- coding: utf-8 -*-
#
# Copyright (c) 2016 Muges
# Copyright (C) 2015 Patrick Griffis <tingping@tingping.se>
# Copyright (C) 2014 Christian Hergert <christian@hergert.me>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
ScrolledWindow that sets a max size for the child to grow into.
"""

from gi.repository import Gtk
from gi.repository import GObject


class ScrolledWindow(Gtk.ScrolledWindow):
    """ScrolledWindow that sets a max size for the child to grow into

    Source
    ------
    https://gist.github.com/TingPing/e70e09722e0fc37d2386
    """
    __gtype_name__ = "EggScrolledWindow"

    max_content_height = GObject.Property(
        type=int, default=-1, nick="Max Content Height",
        blurb="The maximum height request that can be made")
    max_content_width = GObject.Property(
        type=int, default=-1, nick="Max Content Width",
        blurb="The maximum width request that can be made")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.connect_after("notify::max-content-height",
                           lambda obj, param: self.queue_resize())
        self.connect_after("notify::max-content-width",
                           lambda obj, param: self.queue_resize())

    def set_max_content_height(self, value):
        # pylint: disable=arguments-differ
        self.max_content_height = value

    def set_max_content_width(self, value):
        # pylint: disable=arguments-differ
        self.max_content_width = value

    def get_max_content_height(self):
        # pylint: disable=arguments-differ
        return self.max_content_height

    def get_max_content_width(self):
        # pylint: disable=arguments-differ
        return self.max_content_width

    def do_get_preferred_height(self):
        # pylint: disable=arguments-differ
        min_height, natural_height = (
            Gtk.ScrolledWindow.do_get_preferred_height(self)
        )
        child = self.get_child()

        if natural_height and self.max_content_height > -1 and child:
            style = self.get_style_context()
            border = style.get_border(style.get_state())
            additional = border.top + border.bottom

            _, child_nat_height = child.get_preferred_height()
            if all((child_nat_height > natural_height,
                    self.max_content_height > natural_height)):
                natural_height = (
                    min(self.max_content_height, child_nat_height) + additional
                )

        return min_height, natural_height

    def do_get_preferred_width(self):
        # pylint: disable=arguments-differ
        min_width, natural_width = (
            Gtk.ScrolledWindow.do_get_preferred_width(self)
        )
        child = self.get_child()

        if natural_width and self.max_content_width > -1 and child:
            style = self.get_style_context()
            border = style.get_border(style.get_state())
            additional = border.left + border.right + 1

            _, child_nat_width = child.get_preferred_width()
            if all((child_nat_width > natural_width,
                    self.max_content_width > natural_width)):
                natural_width = (
                    min(self.max_content_width, child_nat_width) + additional
                )

        return min_width, natural_width



# -*- coding: utf-8 -*-
#
# Copyright (c) 2016 Muges
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Module handling the conversion of html into plaintext
"""

import logging
import re
import lxml.etree
import lxml.html

BLOCKS = ["p", "br", "div", "h1", "h2", "h3", "h4", "h5", "h6"]
BLOCK_SUFFIX = {
    "br": "\n"
}


class Parser(object):
    """Parser converting html into plaintext"""
    def __init__(self):
        self.logger = logging.getLogger(
            "{}.{}".format(__name__, self.__class__.__name__))

        self.text = ""

    def clear(self):
        """Reinitialize the parser"""
        self.text = ""

    def parse_string(self, html):
        """Parse an html string

        Parameters
        ----------
        html : str
            A html formatted string
        """
        self.clear()

        parser = lxml.etree.HTMLParser(remove_blank_text=True,
                                       remove_comments=True,
                                       remove_pis=True)
        fragments = lxml.html.fragments_fromstring(html, parser=parser)

        if not fragments:
            return ""

        # Add the trailing text
        if isinstance(fragments[0], str):
            text = fragments.pop(0)

            if fragments:
                text = re.sub(r"\s+", " ", text)

            self._add_text(text)

        for node in fragments:
            self._parse_node(node)

        return self.get_text()

    def _parse_node(self, node):
        """Parse an html node

        Parameters
        ----------
        node : lxml.html.HtmlElement
            The current node.
        """
        is_block = node.tag in BLOCKS

        # Handle node's text
        if node.text:
            text = re.sub(r"\s+", " ", node.text)
            if is_block:
                if len(node):  # pylint: disable=len-as-condition
                    text = text.lstrip()
                else:
                    text = text.strip()
            self._add_text(text)

        # Parse children
        for child in node.iterchildren():
            self._parse_node(child)

        if is_block:
            self._rstrip_text()
            self._add_text(BLOCK_SUFFIX.get(node.tag, "\n\n"))

        # Handle node's tail
        if node.tail:
            tail = re.sub(r"\s+", " ", node.tail)
            if is_block:
                tail = tail.strip()
            self._add_text(tail)

    def get_text(self):
        """Return the converted string"""
        return self.text.rstrip(" \t\n")

    def _add_text(self, text):
        """Add text at the end of the output string"""
        try:
            if self.text[-1] == " " and text[0] == " ":
                text = text.rstrip()
        except IndexError:
            pass
        self.text += text

    def _rstrip_text(self):
        """Remove whitespaces at the end of the output string"""
        self.text = self.text.rstrip(" \t\n")


def html_to_plaintext(html):
    """Convert html into plaintext

    Parameters
    ----------
    html : str
        A html formatted string
    """
    if not html:
        return ""

    parser = Parser()
    return parser.parse_string(html)



# -*- coding: utf-8 -*-
#
# Copyright (c) 2016 Muges
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
A Model storing the configuration of the application as ``(key, value)`` pairs.
"""

from collections import namedtuple
import logging
from peewee import TextField

from erika.config import CONFIG_DEFAULTS
from erika.library.fields import JSONField
from .database import BaseModel


class Config(BaseModel):
    """A model storing the configuration of the application as ``(key, value)``
    pairs.

    The keys can be of the form ``group.key``, making it possible to
    get all the values of a certain group at once with the :func:`~get_group`
    method.

    The values can be of any type that is serializable by the ``json`` python
    module."""
    key = TextField(primary_key=True)
    value = JSONField(null=True)

    @classmethod
    def get_value(cls, key):
        """Return the value associated to a key."""
        return cls.get(Config.key == key).value

    @classmethod
    def set_value(cls, key, value):
        """Set the value associated to a key."""
        config, _ = cls.get_or_create(key=key)
        config.value = value
        config.save()

    @classmethod
    def set_defaults(cls):
        """Set the default configuration values (ignoring the values that are
        already defined).

        The default values are stored in the
        :py:obj:`erika.config.CONFIG_DEFAULTS` dictionnary."""
        logger = logging.getLogger(".".join((__name__, cls.__name__)))
        logger.debug("Setting default configuration.")

        (cls
         .insert_many({"key": key, "value": value}
                      for key, value in CONFIG_DEFAULTS.items())
         .on_conflict('IGNORE')
         .execute())

    @staticmethod
    def get_group(group):
        """Return the configurations pairs belonging to a group (i.e. whose
        keys are of the form ``group.*``).

        Example
        -------
        Assuming the config table contains the following data:

        +---------------------+------------------+
        | key                 | value            |
        +=====================+==================+
        | downloads.workers   | 2                |
        +---------------------+------------------+
        | gpodder.synchronize | False            |
        +---------------------+------------------+
        | gpodder.hostname    | 'gpodder.net'    |
        +---------------------+------------------+

        .. code-block:: python

            >>>Config.get_group("downloads")
            configuration(workers=2)
            >>>Config.get_group("gpodder")
            configuration(synchronize=False, hostname='gpodder.net')
            >>>Config.get_group("nothing")
            configuration()

        Parameters
        ----------
        group : str
           The name of the group.

        Returns
        -------
        namedtuple
            A namedtuple containing the ``(key, value)`` pairs.
        """
        prefix = group + "."
        lprefix = len(prefix)
        values = {}

        configs = Config.select().where(Config.key.startswith(prefix))
        for config in configs:
            values[config.key[lprefix:]] = config.value

        type_ = namedtuple("configuration", values.keys())
        return type_(**values)



# -*- coding: utf-8 -*-
#
# Copyright (c) 2016 Muges
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Audio player widgets
"""

from gi.repository import GObject
from gi.repository import Gst
from gi.repository import Gtk
from gi.repository import GLib

from erika.__version__ import __appname__
from erika.util import format_duration
from .player import Player
from .util import cb, get_builder


class PlayerTitle(Gtk.Stack):
    """Widget showing either the applications's name or the current
    episode being played and a progressbar"""
    def __init__(self, player):
        super().__init__()
        self.player = player
        self.seeking = False

        builder = get_builder('data/player_title.glade')
        builder.connect_signals(self)

        player = builder.get_object("player")
        loading = builder.get_object("loading")
        title = Gtk.Label()
        title.set_markup("<b>{}</b>".format(__appname__))

        self.add_named(title, "title")
        self.add_named(player, "player")
        self.add_named(loading, "loading")

        self.playing = [
            builder.get_object("playing_1"),
            builder.get_object("playing_2")
        ]
        self.progress = builder.get_object("progress")
        self.position = builder.get_object("position")
        self.duration = builder.get_object("duration")

        self.show_all()
        self.set_player_state(None, Player.STOPPED)

    def set_player_state(self, episode, state):
        """Called when the state of the player changes

        Parameters
        ----------
        episode : Episode
            The episode currently being played.
        state : int
            The state of the player (Player.BUFFERING, Player.STOPPED,
            Player.PAUSED, or Player.PLAYING)
        """
        if episode:
            for playing in self.playing:
                playing.set_markup(
                    "<b>{}</b> from <b><i>{}</i></b>".format(
                        GLib.markup_escape_text(episode.title),
                        GLib.markup_escape_text(episode.podcast.title)))

        if state in [Player.PLAYING, Player.PAUSED]:
            self._update_progress()
            self.set_visible_child_name('player')
        elif state == Player.BUFFERING:
            self.set_visible_child_name('loading')
        elif state == Player.STOPPED:
            self.set_visible_child_name('title')
        else:
            raise ValueError("Invalid player state : {}.".format(state))

    def set_progress(self, position, duration):
        """Called when the progress of the playback changes"""
        self.progress.set_range(0, duration)
        if not self.seeking:
            self.progress.set_value(position)
        self.progress.set_fill_level(self.player.get_buffered())

        self.position.set_text(format_duration(position // Gst.SECOND))
        self.duration.set_text(format_duration(duration // Gst.SECOND))

    def _on_seeking_start(self, scale, event):
        """
        Called when the user starts seeking
        """
        if event.button == 1:
            self.seeking = True

    def _on_seeking_end(self, scale, event):
        """
        Called when the user ends seeking
        """
        if event.button == 1:
            self.player.seek(self.progress.get_value())
            self.seeking = False

    def _update_progress(self):
        """
        Update the slider
        """
        position = self.player.get_position()
        duration = self.player.get_duration()

        self.set_progress(position, duration)


class PlayerControls(Gtk.Box):
    """Widget containing buttons to control the player"""
    def __init__(self, player):
        super().__init__()
        self.set_orientation(Gtk.Orientation.HORIZONTAL)
        self.set_no_show_all(True)

        self.player = player

        backward = Gtk.Button.new_from_icon_name(
            "media-seek-backward-symbolic", Gtk.IconSize.SMALL_TOOLBAR)
        backward.connect("clicked", cb(self.player.seek_relative),
                         -30 * Gst.SECOND)
        backward.show()

        self.play = Gtk.Button.new_from_icon_name(
            "media-playback-start-symbolic", Gtk.IconSize.SMALL_TOOLBAR)
        self.play.connect("clicked", cb(self.player.play))

        self.pause = Gtk.Button.new_from_icon_name(
            "media-playback-pause-symbolic", Gtk.IconSize.SMALL_TOOLBAR)
        self.pause.connect("clicked", cb(self.player.pause))

        stop = Gtk.Button.new_from_icon_name(
            "media-playback-stop-symbolic", Gtk.IconSize.SMALL_TOOLBAR)
        stop.connect("clicked", cb(self.player.stop))
        stop.show()

        forward = Gtk.Button.new_from_icon_name(
            "media-seek-forward-symbolic", Gtk.IconSize.SMALL_TOOLBAR)
        forward.connect("clicked", cb(self.player.seek_relative),
                        30 * Gst.SECOND)
        forward.show()

        self.pack_start(backward, False, True, 0)
        self.pack_start(self.play, False, True, 0)
        self.pack_start(self.pause, False, True, 0)
        self.pack_start(stop, False, True, 0)
        self.pack_start(forward, False, True, 0)

        self.set_player_state(Player.STOPPED)

    def set_player_state(self, state):
        """Called when the state of the player changes

        Parameters
        ----------
        state : int
            The state of the player (Player.BUFFERING, Player.STOPPED,
            Player.PAUSED, or Player.PLAYING)
        """
        if state in [Player.PLAYING, Player.PAUSED]:
            self.pause.set_visible(state == Player.PLAYING)
            self.play.set_visible(state == Player.PAUSED)
            self.show()
        elif state == Player.BUFFERING:
            self.hide()
        elif state == Player.STOPPED:
            self.hide()
        else:
            raise ValueError("Invalid player state : {}.".format(state))



# -*- coding: utf-8 -*-
#
# Copyright (c) 2016 Muges
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
The :mod:`erika.library` module implements the podcast library.
"""

import logging
import os

from playhouse.sqlite_ext import SqliteExtDatabase

from .models import (database, Config, Episode, EpisodeAction, Podcast,
                     PodcastAction)
from erika.__version__ import __version_tuple__


def initialize(configuration_directory):
    """Create or update the database

    This function should be called once at the beginning of the
    program. If the database file does not exists, it creates it. If
    the database is outdated (e.g. if it was created by an old version
    of the program, and the database format has changed), it should be
    updated.
    """
    logger = logging.getLogger(__name__)
    logger.debug("Initializing the database")

    database_path = os.path.join(configuration_directory, "library")
    database.init(database_path)

    with database.transaction():
        if not Config.table_exists():
            # This is the first time the application is launched, set the
            # version to prevent the database migration from happening
            Config.create_table()
            Config.set_value("application.version", __version_tuple__)

        # Create the other tables
        for model in [Config, Podcast, Episode, EpisodeAction, PodcastAction]:
            if not model.table_exists():
                model.create_table()

        # Set default configuration values
        Config.set_defaults()

        # Migrate database
        version = tuple(Config.get_value("application.version"))
        if version < __version_tuple__:
            migrate(version)

        # Set current version
        Config.set_value("application.version", __version_tuple__)

        # Set all the episodes as non-new
        Episode.update(new=False)


def migrate(version):
    """Migrate the database."""
    logger = logging.getLogger(__name__)
    logger.debug("Migrating the database")

    if version < (0, 1, 1):
        logger.debug("Migrating to version 0.1.1.")

        # Paths were previously store as absolute paths, convert to paths
        # relative to the root of the library
        with database.transaction():
            episodes = Episode.select().where(Episode.local_path.is_null(False))
            for episode in episodes:
                episode.absolute_local_path = episode.local_path
                episode.save()


def scan():
    """Scan the library directory to add the local episode files."""
    logger = logging.getLogger(__name__)
    logger.info("Scanning the library")

    with database.transaction():
        # Check that the files that are in the database still exist
        # and update the `paths` list so that it contains all the
        # files that are already the database
        paths = []
        episodes = Episode.select().where(Episode.local_path.is_null(False))
        for episode in episodes:
            if os.path.isfile(episode.absolute_local_path):
                paths.append(episode.absolute_local_path)
            else:
                episode.local_path = None
                episode.save()

        # Look for new local files
        library_root = Config.get_value("library.root")
        for dirpath, _, filenames in os.walk(library_root):
            for filename in filenames:
                path = os.path.join(dirpath, filename)
                if path in paths:
                    # The file is already in the library
                    continue

                if path.endswith(".part"):
                    # The file is incomplete
                    continue

                episode = Episode.get_matching(path)
                if not episode:
                    # No match was found
                    continue

                # Set the path of the episode in the database
                episode.absolute_local_path = path
                episode.save()

                # Add an action indicating that the episode has been
                # downloaded on this device
                action = EpisodeAction(episode=episode, action="download")
                action.save()



# -*- coding: utf-8 -*-
#
# Copyright (c) 2016 Muges
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Module handling the conversion of plaintext into html
"""

import html
import re
import string

SPACES_PATTERN = re.compile(r"[ \t]+")
NEWLINE_PATTERN = re.compile(r"[ \t]*\n[ \t]*")
PARAGRAPH_PATTERN = re.compile(r"[ \t]*\n[ \t]*\n[ \t]*")
URL_PATTERN = re.compile(
    # Protocol
    r"(http|ftp|https)://"
    # Domain
    r"([\w_-]+(?:(?:\.[\w_-]+)+))"
    # Path
    r"([\w.,@?^=%&:;/~+#-]*[\w@?^=%&/~+#-])?"
)


def paragraph_to_html(paragraph):
    """Convert a paragraph to html"""
    paragraph = paragraph.strip(string.whitespace)
    paragraph = html.escape(paragraph)

    # Line breaks
    paragraph = NEWLINE_PATTERN.sub("<br/>\n", paragraph)

    # Spaces
    paragraph = SPACES_PATTERN.sub(" ", paragraph)

    # Urls
    paragraph = URL_PATTERN.sub(r'<a href="\g<0>">\g<0></a>', paragraph)

    return "<p>{}</p>".format(paragraph)


def plaintext_to_html(plaintext):
    """Convert a text to html"""
    paragraphs = PARAGRAPH_PATTERN.split(plaintext)

    return "\n".join([
        paragraph_to_html(paragraph)
        for paragraph in paragraphs
        if not PARAGRAPH_PATTERN.match(paragraph)
    ])



# -*- coding: utf-8 -*-
#
# Copyright (c) 2016 Muges
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
A ListBox whose rows are indexed
"""

import logging

from .listbox import ListBox


class IndexedListBox(ListBox):
    """A ListBox whose rows are indexed"""
    def __init__(self):
        super().__init__()
        self.logger = logging.getLogger(
            "{}.{}".format(__name__, self.__class__.__name__))

        self.rows = {}

    def get_ids(self):
        """Return the ids of all the rows

        Returns
        -------
        set
        """
        return set(self.rows.keys())

    def get_row(self, row_id):
        """Return a row given its id

        Parameters
        ----------
        row_id
            The id of the row

        Returns
        -------
        Gtk.ListBoxRow

        Raises
        ------
        ValueError if there is no row with this id
        """
        if row_id not in self.rows:
            raise ValueError(
                "The ListBox has no row with id '{}'.".format(row_id))

        return self.rows[row_id]

    def add_with_id(self, row, row_id):
        """Add a row with a certain id

        Parameters
        ----------
        row : Gtk.ListBoxRow
            The row
        row_id
            The id of the row

        Raises
        ------
        ValueError if there is already a row with this id
        """
        if row_id in self.rows:
            raise ValueError(
                "The ListBox already has a row with id '{}'.".format(row_id))

        self.rows[row_id] = row
        self.add(row)

    def remove_id(self, row_id):
        """Remove a row given its id

        Parameters
        ----------
        row_id
            The id of the row to remove

        Raises
        ------
        ValueError if there is no row with this id
        """
        try:
            row = self.rows.pop(row_id)
        except KeyError:
            raise ValueError(
                "The ListBox has no row with id '{}'.".format(row_id))

        self.remove(row)

    def clear(self):
        """Remove all the rows"""
        for child in self.get_children():
            child.destroy()

        self.rows = {}



# -*- coding: utf-8 -*-
#
# Copyright (c) 2016 Muges
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
A Gtk.ListBox patched to handle multiple selection
"""

from gi.repository import Gtk
from gi.repository import Gdk


class ListBox(Gtk.ListBox):
    """A Gtk.ListBox patched to handle multiple selection"""
    def __init__(self):
        Gtk.ListBox.__init__(self)
        self.connect("button-press-event", ListBox._on_button_press_event)
        self.connect("key-press-event", ListBox._on_key_press_event)

        self.last_clicked = None
        self.last_focused = None
        self.filter_func = None

    def remove(self, child):  # pylint: disable=arguments-differ
        """Remove a child element"""
        if self.last_clicked == child:
            self.last_clicked = None
        if self.last_focused == child:
            self.last_focused = None
        Gtk.ListBox.remove(self, child)

    def set_filter_func(self, filter_func):  # pylint: disable=arguments-differ
        # Keep a references to the filter function
        self.filter_func = filter_func
        Gtk.ListBox.set_filter_func(self, filter_func)

    def invalidate_filter(self):  # pylint: disable=arguments-differ
        Gtk.ListBox.invalidate_filter(self)

        if self.last_clicked and not self.filter_func(self.last_clicked):
            self.last_clicked = None
        if self.last_focused and not self.filter_func(self.last_focused):
            self.last_focused = None

        # Unselect the rows that are not visible
        for row in self.get_selected_rows():
            if not self.filter_func(row):
                self.unselect_row(row)

    def _on_button_press_event(self, event):
        # Propagate the event if it is not a single left button click or in
        # single selection mode
        if any((self.get_selection_mode() != Gtk.SelectionMode.MULTIPLE,
                event.type != Gdk.EventType.BUTTON_PRESS,
                event.button != 1 and event.button != 3)):
            return False

        # Get the row that was clicked
        row = self.get_row_at_y(event.y)

        if event.button == 1:
            # Left click
            if row is None:
                return True

            if self.last_clicked is None:
                self.last_clicked = self.get_row_at_index(0)

            if event.state & Gdk.ModifierType.SHIFT_MASK:
                self._select_interval(row)
            elif event.state & Gdk.ModifierType.CONTROL_MASK:
                self._add_to_selection(row)
            else:
                self._select_single(row)

            return True
        elif event.button == 3:
            # Right click
            if row and not row.is_selected():
                self._select_single(row)

            self.emit("popup-menu")

            return False

    def _on_key_press_event(self, event):
        # Propagate the event if the key pressed is not up or down or in single
        # selection mode
        if any((self.get_selection_mode() != Gtk.SelectionMode.MULTIPLE,
                event.keyval != Gdk.KEY_Up and event.keyval != Gdk.KEY_Down)):
            return False

        if self.last_clicked is None:
            self.last_clicked = self.get_row_at_index(0)
        if self.last_focused is None:
            self.last_focused = self.get_row_at_index(0)

        # Get the index of the next or previous row depending on the key
        # pressed
        index = self.last_focused.get_index()
        if event.keyval == Gdk.KEY_Up:
            delta = 1
        elif event.keyval == Gdk.KEY_Down:
            delta = 1

        # Get the next/previous visible row
        while True:
            index += delta
            row = self.get_row_at_index(index)
            if row is None:
                return True
            if self.filter_func(row):
                break

        if event.state & Gdk.ModifierType.SHIFT_MASK:
            self._select_interval(row)
        else:
            self._select_single(row)

        return True

    def _add_to_selection(self, row):
        """Add the row to the current selection

        Parameters
        ----------
        row : Gtk.ListBoxRow
            A child of the ListBox
        """
        self.select_row(row)
        self.last_clicked = row
        self.last_focused = row
        row.grab_focus()

    def _select_single(self, row):
        """Add the row to the current selection

        Parameters
        ----------
        row : Gtk.ListBoxRow
            A child of the ListBox
        """
        # Select only this row
        self.unselect_all()
        self.select_row(row)
        self.last_clicked = row
        self.last_focused = row
        row.grab_focus()

    def _select_interval(self, row):
        """Select the all the rows between the one that was clicked last and
        the one in the parameters

        Parameters
        ----------
        row : Gtk.ListBoxRow
            A child of the ListBox
        """
        self.unselect_all()
        self.last_focused = row
        row.grab_focus()

        first = self.last_clicked.get_index()
        last = row.get_index()
        if first > last:
            first, last = last, first

        for index in range(first, last + 1):
            row = self.get_row_at_index(index)
            # Select only the visible rows
            if self.filter_func(row):
                self.select_row(row)



# -*- coding: utf-8 -*-
#
# Copyright (c) 2016 Muges
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
The :mod:`erika.library.gpodder` module handles the synchronization of the
library with gpodder.net.
"""

import logging
from mygpoclient.api import MygPodderClient
from mygpoclient.simple import MissingCredentials
from mygpoclient.http import Unauthorized
from mygpoclient.util import iso8601_to_datetime

from .models import (database, Config, Episode, EpisodeAction, Podcast,
                     PodcastAction)


class GPodderUnauthorized(Exception):
    """Exception raised when the clients tries to connect to gpodder.net with
    invalid credentials."""

    def __init__(self):
        super().__init__("Invalid credentials. Unable to connect to "
                         "gpodder.net.")


class RemoteEpisodeAction(object):  # pylint: disable=too-few-public-methods
    """Object representing a remote episode action.

    It has the same attributes as an
    :class:`~erika.library.models.EpisodeAction`.
    """
    def __init__(self, action):
        self.podcast_url = action.podcast
        self.episode_url = action.episode,
        self.action = action.action
        self.time = iso8601_to_datetime(action.timestamp)
        self.started = action.started
        self.position = action.position
        self.total = action.total

    def __str__(self):
        attrs = ["podcast_url", "episode_url", "action", "time", "started",
                 "position", "total"]
        formatted_attrs = ["{}={}".format(attr, repr(getattr(self, attr)))
                           for attr in attrs]
        return "RemoteEpisodeAction({})".format(", ".join(formatted_attrs))


class GPodderClient(object):
    """Client handling the synchronization of the library with gpodder.net."""
    def __init__(self):
        self.logger = logging.getLogger(
            ".".join((__name__, self.__class__.__name__)))

        self.configuration = Config.get_group("gpodder")
        self.smart_mark_seconds = Config.get_value("player.smart_mark_seconds")

        self.client = None

    def enabled(self, connecting=False):
        """Return True if the synchronization is enabled and the client is
        connected.

        Parameters
        ----------
        connecting : bool, optional
            True to return True even if the client is not connected yet
            (default is False).
        """
        return all((self.configuration.synchronize,
                    self.configuration.username,
                    self.configuration.password,
                    self.configuration.hostname,
                    self.client or connecting))

    def connect(self):
        """Connect to gpodder.net.

        Returns
        -------
        bool
            True if the connection attempt was successful.
        """
        if not self.enabled(connecting=True):
            return False

        self.logger.debug("Connecting to gpodder.net.")
        try:
            self.client = MygPodderClient(self.configuration.username,
                                          self.configuration.password,
                                          self.configuration.hostname)
            self.update_device()
        except (MissingCredentials, Unauthorized):
            self.logger.warning('Invalid credentials. Unable to connect to '
                                'gpodder.net.')
            self.client = None
            raise GPodderUnauthorized()
        except Exception:  # pylint: disable=broad-except
            self.logger.exception("Unable to connect to gpodder.net.")
            self.client = None
            raise

        return self.client is not None

    def update_device(self):
        """Update the device name."""
        if not self.enabled():
            return

        if self.configuration.devicename_changed:
            # The device name changed locally, send it to gpodder.net
            self.client.update_device_settings(self.configuration.deviceid,
                                               self.configuration.devicename)
            Config.set_value("gpodder.devicename_changed", False)
        else:
            # The device name was not changed locally, get it from gpodder.net
            devices = self.client.get_devices()
            for device in devices:
                if device.device_id == self.configuration.deviceid:
                    Config.set_value("gpodder.devicename", device.caption)
                    break

        self.configuration = Config.get_group("gpodder")

    def pull_subscriptions(self):
        """Pull the subscriptions changes from gpodder.net and process them."""
        if not self.enabled():
            return

        self.logger.debug("Pulling subscription changes.")

        try:
            changes = self.client.pull_subscriptions(
                self.configuration.deviceid,
                self.configuration.last_subscription_sync)
        except Exception:  # pylint: disable=broad-except
            self.logger.exception("Unable to pull subscriptions from "
                                  "gpodder.net.")
            return

        with database.transaction():
            # Apply the changes
            for url in changes.remove:
                self.logger.debug("Removing podcast %s.", url)
                (Podcast
                 .delete()
                 .where(Podcast.url == url)
                 .execute())

            for url in changes.add:
                self.logger.debug("Adding podcast %s.", url)
                Podcast.new("rss", url)

            Config.set_value("gpodder.last_subscription_sync", changes.since)

    def push_subscriptions(self):
        """Push the local subscriptions changes to gpodder.net.

        Returns
        -------
        datetime, optional
            The time at which the subscriptions were synchronized.
        """
        if not self.enabled():
            return

        self.logger.debug("Pushing subscription changes.")

        # Get the podcasts that were added or removed since the last
        # synchronization
        add = PodcastAction().select().where(
            PodcastAction.action == 'add'
        )
        added_urls = [action.podcast_url for action in add]

        remove = PodcastAction().select().where(
            PodcastAction.action == 'remove'
        )
        removed_urls = [action.podcast_url for action in remove]

        # Send them to gpodder.net
        if added_urls or removed_urls:
            try:
                result = self.client.update_subscriptions(
                    self.configuration.deviceid, added_urls, removed_urls)
            except Exception:  # pylint: disable=broad-except
                self.logger.exception("Unable to push subscriptions to "
                                      "gpodder.net.")
                return

            with database.transaction():
                # Update the podcasts urls
                for old_url, new_url in result.update_urls:
                    if new_url:
                        self.logger.debug("Updating url '%s' to '%s'.",
                                          old_url, new_url)
                        (Podcast
                         .update(url=new_url)
                         .where(Podcast.url == old_url)
                         .on_conflict("REPLACE")
                         .execute())

                # Remove the actions
                for action in add:
                    action.delete_instance()
                for action in remove:
                    action.delete_instance()

            return result.since

        return None

    def synchronize_subscriptions(self):
        """Synchronize the subscriptions changes with gpodder.net."""
        if not self.enabled():
            return

        self.pull_subscriptions()
        since = self.push_subscriptions()

        if since is not None:
            Config.set_value("gpodder.last_subscription_sync", since)

    def process_episode_actions(self, actions):
        """Process the episodes actions.

        Parameters
        ----------
        actions : list of :class:`EpisodeAction` and :class:`RemoteEpisodeAction`
            A list containing all the local and remote episode actions.
        since : int
            The time at which the remote actions were pulled.
        """
        actions.sort(key=lambda a: a.time)

        with database.transaction():
            for action in actions:
                try:
                    self.process_episode_action(action)
                except Exception:  # pylint: disable=broad-except
                    self.logger.exception("Unable to process action %s",
                                          action)

    def process_episode_action(self, action):
        """Process an episode action."""
        if action.action == "play":
            if action.position + self.smart_mark_seconds >= action.total:
                (Episode
                 .update(played=True, progress=0)
                 .where(Episode.file_url == action.episode_url)
                 .execute())
            else:
                (Episode
                 .update(progress=action.position)
                 .where(Episode.file_url == action.episode_url)
                 .execute())
        elif action.action == "new":
            (Episode
             .update(played=False)
             .where(Episode.file_url == action.episode_url)
             .execute())

    def pull_episode_actions(self):
        """Pull episode actions from gpodder.net."""
        if not self.enabled():
            return

        self.logger.debug("Pulling episode actions.")

        try:
            changes = self.client.download_episode_actions(
                self.configuration.last_episodes_sync)
        except Exception:  # pylint: disable=broad-except
            self.logger.exception("Unable to pull episode actions from "
                                  "gpodder.net.")
            return

        local_actions = [action for action in EpisodeAction.select()]
        remote_actions = [RemoteEpisodeAction(action)
                          for action in changes.actions]
        self.process_episode_actions(local_actions + remote_actions)

        Config.set_value("gpodder.last_episodes_sync", changes.since)

    def push_episode_actions(self):
        """Push episode actions to gpodder.net."""
        if not self.enabled():
            return

        self.logger.debug("Pushing episode actions.")

        actions = [action for action in EpisodeAction.select()]

        try:
            since = self.client.upload_episode_actions(
                [action.for_gpodder() for action in actions])
        except Exception:  # pylint: disable=broad-except
            self.logger.exception("Unable to push episode actions to "
                                  "gpodder.net.")
            return

        # Remove actions
        with database.transaction():
            for action in actions:
                action.delete_instance()

        return since

    def synchronize_episode_actions(self):
        """Synchronize episode actions with gpodder.net."""
        if not self.enabled():
            return

        self.pull_episode_actions()
        since = self.push_episode_actions()

        if since is not None:
            Config.set_value("gpodder.last_episodes_sync", since)



# -*- coding: utf-8 -*-
#
# Copyright (c) 2016 Muges
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
List of podcasts
"""

import locale

from gi.repository import Gtk
from gi.repository import GObject
from gi.repository import GLib
from gi.repository import Gio

from erika.library.models import Podcast
from .util import cb
from .widgets import Label, IndexedListBox

IMAGE_SIZE = 64
SUBTITLE_LINES = 2


def sort_func(row1, row2):
    """Compare two rows to determine which should be first. Sort them
    alphabetically by podcast title.

    Parameters
    ----------
    row1 : Gtk.ListBoxRow
        The first row
    row2 : Gtk.ListBoxRow
        The second row

    Returns
    -------
    int
        -1 if row1 should be before row2,
        0 if they are equal,
        1 otherwise
    """
    if row1.podcast.title is None:
        return 1
    elif row2.podcast.title is None:
        return -1
    else:
        return locale.strcoll(row1.podcast.title, row2.podcast.title)


class PodcastList(Gtk.VBox):
    """
    List of podcasts

    Signals
    -------
    podcast-selected(Podcast)
        Emitted when a podcast is selected
    """

    __gsignals__ = {
        'podcast-selected':
            (GObject.SIGNAL_RUN_FIRST, None, (GObject.TYPE_PYOBJECT,)),
    }

    def __init__(self):
        super().__init__()

        # Podcast list
        self.list = IndexedListBox()
        self.list.set_selection_mode(Gtk.SelectionMode.BROWSE)
        self.list.set_sort_func(sort_func)
        self.list.connect("row-selected", self._on_row_selected)

        scrolled_window = Gtk.ScrolledWindow()
        scrolled_window.add_with_viewport(self.list)

        # Action bar
        self.action_bar = Gtk.ActionBar()

        self.add_podcast = Gtk.Button.new_from_icon_name(
            "list-add-symbolic", Gtk.IconSize.SMALL_TOOLBAR)
        self.add_podcast.set_tooltip_text("Add a podcast")
        self.add_podcast.set_action_name("app.add-podcast")
        self.action_bar.pack_start(self.add_podcast)

        remove_podcast = Gtk.Button.new_from_icon_name(
            "list-remove-symbolic", Gtk.IconSize.SMALL_TOOLBAR)
        remove_podcast.set_tooltip_text("Remove the selected podcast")
        remove_podcast.connect("clicked", self._on_remove_podcast)
        self.action_bar.pack_start(remove_podcast)

        import_opml = Gtk.Button.new_from_icon_name(
            "document-open-symbolic", Gtk.IconSize.SMALL_TOOLBAR)
        import_opml.set_action_name("app.import-opml")
        import_opml.set_tooltip_text("Import an opml file")
        self.action_bar.pack_end(import_opml)

        export_opml = Gtk.Button.new_from_icon_name(
            "document-save-as-symbolic", Gtk.IconSize.SMALL_TOOLBAR)
        export_opml.set_action_name("app.export-opml")
        export_opml.set_tooltip_text("Export subscriptions to an opml file")
        self.action_bar.pack_end(export_opml)

        # Layout
        self.pack_start(scrolled_window, True, True, 0)
        self.pack_start(self.action_bar, False, False, 0)

        application = Gio.Application.get_default()
        application.connect('network-state-changed',
                            cb(self.set_network_state))

    def update(self):
        """Update the list of podcasts"""
        # Set of ids that should be removed
        remove_ids = self.list.get_ids()

        for podcast in Podcast.select():
            try:
                row = self.list.get_row(podcast.id)
            except ValueError:
                # The row does not exist, create it
                row = PodcastRow(podcast)
                self.list.add_with_id(row, podcast.id)
            else:
                # The row already exists, update it
                remove_ids.remove(podcast.id)
                row.podcast = podcast
                row.update()

        # Remove the podcasts that are not in the database anymore
        for podcast_id in remove_ids:
            self.list.remove_id(podcast_id)

        if self.list.get_selected_row() is None:
            self.list.select_row(self.list.get_row_at_index(0))

        self.list.invalidate_sort()

    def update_podcast(self, podcast):
        """Update a podcast

        Parameters
        ----------
        podcast_id : int
            The id of the podcast
        """
        try:
            row = self.list.get_row(podcast.id)
        except ValueError:
            # The row does not exist, create it
            row = PodcastRow(podcast)
            self.list.add_with_id(row, podcast.id)
        else:
            row.podcast = podcast
            row.update()

    def update_current(self):
        """Update the selected row"""
        row = self.list.get_selected_row()

        if row:
            row.update()

    def _on_row_selected(self, listbox, row):
        """Called when a row is selected.

        Emit the podcast-selected signal.
        """
        if row:
            self.emit('podcast-selected', row.podcast)

    def _on_remove_podcast(self, button):
        selection = self.list.get_selected_row()

        if selection is None:
            return

        # Get parent window
        window = self.get_toplevel()

        # Ask for confirmation
        dialog = Gtk.MessageDialog(window, 0, Gtk.MessageType.QUESTION,
                                   Gtk.ButtonsType.YES_NO)
        dialog.set_markup(
            "Are you sure you want to remove the podcast <b>{}</b>?".format(
                GLib.markup_escape_text(selection.podcast.display_title)))
        response = dialog.run()
        dialog.destroy()

        if response == Gtk.ResponseType.NO:
            return

        # Delete the podcast
        selection.podcast.delete_instance(recursive=True)
        selection.destroy()

        # Select the first row
        first_row = self.list.get_row_at_index(0)
        self.list.select_row(first_row)

    def set_network_state(self, online, _):
        """Called when the network state is changed"""
        self.add_podcast.set_sensitive(online)


class PodcastRow(Gtk.ListBoxRow):
    """Row in the list of podcasts"""
    def __init__(self, podcast):
        Gtk.ListBoxRow.__init__(self)

        self.podcast = podcast

        self.grid = Gtk.Grid()
        self.grid.set_column_spacing(5)
        self.add(self.grid)

        # Podcast Image
        self.icon = Gtk.Image()
        self.grid.attach(self.icon, 0, 0, 1, 2)

        # Podcast title
        self.title = Label()
        self.title.set_hexpand(True)
        self.grid.attach(self.title, 1, 0, 1, 1)

        # Unplayed and new counts
        self.counts = Label()
        self.counts.set_alignment(xalign=1, yalign=0.5)
        self.grid.attach(self.counts, 2, 0, 2, 1)

        # Podcast subtitle
        self.subtitle = Label(lines=SUBTITLE_LINES)
        self.grid.attach(self.subtitle, 1, 1, 3, 1)

        self.error_icon = Gtk.Image.new_from_icon_name(
            'gtk-dialog-error', Gtk.IconSize.BUTTON)
        self.error_icon.set_tooltip_text(
            'The last update of this podcast failed.')

        self.update()

    def update(self):
        """Update the widget"""
        self.grid.set_tooltip_markup((
            "<b>{title}</b>\n"
            "{episodes} episodes"
        ).format(
            title=GLib.markup_escape_text(self.podcast.display_title),
            episodes=self.podcast.episodes_count
        ))

        # Podcast Image
        self.icon.set_from_pixbuf(self.podcast.image.as_pixbuf(IMAGE_SIZE))

        # Podcast title
        self.title.set_markup("<b>{}</b>".format(
            GLib.markup_escape_text(self.podcast.display_title)))

        # Unplayed and new counts
        if self.podcast.unplayed_count > 0:
            unplayed = str(self.podcast.unplayed_count)
        else:
            unplayed = ""

        if self.podcast.new_count > 0:
            new = "<b>({})</b>".format(self.podcast.new_count)
        else:
            new = ""

        if unplayed and new:
            self.counts.set_markup(unplayed + " " + new)
        else:
            self.counts.set_markup(unplayed + " " + new)

        # Update layout if there was an error
        self.grid.remove(self.counts)
        self.grid.remove(self.error_icon)
        if self.podcast.update_failed:
            self.grid.attach(self.counts, 2, 0, 1, 1)
            self.grid.attach(self.error_icon, 3, 0, 1, 1)
        else:
            self.grid.attach(self.counts, 2, 0, 2, 1)

        # Podcast subtitle (only the first line)
        self.subtitle.set_text(
            self.podcast.display_subtitle.split('\n')[0])

        self.show_all()



# -*- coding: utf-8 -*-
#
# Copyright (c) 2016 Muges
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
A field used to store an image.
"""

import base64
from io import BytesIO
from PIL import Image as PImage
from peewee import BlobField

from gi.repository import GdkPixbuf


class ImageField(BlobField):
    """A field used to store an image."""
    def db_value(self, image):
        return image.data

    def python_value(self, data):
        return Image(data)


class Image(object):
    """
    Object representing an image.

    If an Image is used in a format string, it will be returned as a data uri,
    making it possible to embed it in html as follow::

        '<img src="{image}" />'.format(image=image)
    """
    def __init__(self, data):
        self.data = data

    def __bool__(self):
        return self.data is not None

    def get_data(self, width=None):
        """
        Return a byte sequence containing the image in png format.

        Parameters
        ----------
        width : int, optional
            The width of the image, or None to keep the original width.
        """
        if self.data:
            image = PImage.open(BytesIO(self.data))
        else:
            image = PImage.new('RGBA', (1, 1), (0, 0, 0, 0))

        if width:
            image = image.resize(
                (width, width * image.height // image.width),
                PImage.LANCZOS)

        data = BytesIO()
        image.save(data, format='PNG')
        return data.getvalue()

    def as_pixbuf(self, width=None):
        """
        Return the image as :class:`GdkPixbuf.Pixbuf`.

        Parameters
        ----------
        width : int, optional
            The width of the pixbuf, or None to keep the original width.
        """
        loader = GdkPixbuf.PixbufLoader.new_with_mime_type("image/png")
        loader.write(self.get_data(width))
        loader.close()
        pixbuf = loader.get_pixbuf()

        return pixbuf

    def __format__(self, spec=None):
        """
        Return the data uri corresponding to the image.

        Parameters
        ----------
        spec : str, optional
            The width of the image, or None to keep the original width.
        """
        if spec:
            width = int(spec)
        else:
            width = None

        return "data:image/png;base64,{}".format(
            base64.b64encode(self.get_data(width)).decode())



# -*- coding: utf-8 -*-
#
# Copyright (c) 2016 Muges
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Utility functions
"""

from itertools import chain
import mimetypes
import re

EXTENSIONS = {
    "audio/mpeg": ".mp3",
    "audio/x-m4a": ".m4a",
    "video/mp4": ".mp4",
    "video/x-m4v": ".m4v",
    "video/quicktime": ".mov",
    "application/pdf": ".pdf",
    "document/x-epub": ".epub",
}

DELETE_CHARS = (
    ''.join(map(chr, chain(range(0, 32), range(127, 160)))) +
    '?<>|"'
)

TRANSLATION_TABLE = str.maketrans('/\\:*\n\t', '----  ', DELETE_CHARS)

MAX_FILENAME_LENGTH = 255

# List of filenames that are reserved on linux or windows
FORBIDDEN_FILENAMES = [
    ".", "..", "CON", "PRN", "AUX", "NUL", "COM1", "COM2", "COM3", "COM4",
    "COM5", "COM6", "COM7", "COM8", "COM9", "LPT1", "LPT2", "LPT3", "LPT4",
    "LPT5", "LPT6", "LPT7", "LPT8", "LPT9"
]

HYPHEN_PATTERN = r"(\S)(\s+-|-\s+|\s+-\s+)(\S)"
HYPHEN_REPLACE = r"\1 - \3"


def guess_extension(mimetype):
    """Guess a file's extension from its mimetype.

    The :func:`mimetypes.guess_extension` function sometimes return uncommon
    extensions, such as ``.mp2`` instead of ``.mp3`` for mimetype
    ``audio/mpeg``. This function tries to return consistent extensions for
    common podcast formats, and falls back to :func:`mimetypes.guess_extension`
    for other formats.

    Parameters
    ----------
    mimetype : str
        The mimetype of a file.

    Returns
    -------
    str
        The extension of the file, with a trailing '.'
    """
    try:
        return EXTENSIONS[mimetype]
    except KeyError:
        return mimetypes.guess_extension(mimetype) or ''


def sanitize_filename(filename):
    """Sanitize a string to make it a valid filename.

    This should hopefully work for windows and linux.
    """
    # Remove or replace forbidden characters
    filename = filename.translate(TRANSLATION_TABLE)

    # Make sure hyphens are surrounded by exactly one space on each side or no
    # spaces at all
    filename = re.sub(HYPHEN_PATTERN, HYPHEN_REPLACE, filename)

    # Truncate filenames
    filename = filename[:MAX_FILENAME_LENGTH]

    if filename in FORBIDDEN_FILENAMES:
        filename += "-podcast"

    return filename



# -*- coding : utf-8 -*-
#
# Copyright (c) 2016 Muges
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os
import glob
import pytest

from podcasts.util.html_to_plaintext import html_to_plaintext

DATA_DIRECTORY = "tests/util/data/html_to_plaintext"

PATTERN = os.path.join(DATA_DIRECTORY, "*.txt")
FILES = glob.glob(PATTERN)
NAMES = [os.path.splitext(path)[0] for path in FILES]

@pytest.mark.parametrize("filename", NAMES)
def test_html_to_plaintext(filename):
    with open(filename + ".txt") as fileobj:
        plaintext = fileobj.read().rstrip()
    with open(filename + ".html") as fileobj:
        html = fileobj.read()

    assert html_to_plaintext(html) == plaintext



# -*- coding: utf-8 -*-
#
# Copyright (c) 2016 Muges
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
GTK frontend
"""

import sys
import gi

gi.require_version('Gtk', '3.0')
gi.require_version('Gst', '1.0')
gi.require_version('WebKit', '3.0')
gi.require_version('GdkPixbuf', '2.0')

# pylint: disable=wrong-import-position
from gi.repository import Gst
from gi.repository import GObject

from .application import Application
# pylint: enable=wrong-import-position


def run():
    """
    Start the application
    """
    # Needed for PyGobject < 3.10.2
    GObject.threads_init()
    Gst.init_check(None)

    app = Application()
    app.run(sys.argv)



# -*- coding: utf-8 -*-
#
# Copyright (c) 2016 Muges
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Widget displaying the details of a podcast or episode
"""

import threading
import pkgutil

from gi.repository import Gtk
from gi.repository import GObject
from gi.repository import Gio

from erika.frontend.widgets import WebView


class Details(Gtk.ScrolledWindow):
    """Widget displaying the details of a podcast or episode"""
    EPISODE_TEMPLATE = pkgutil.get_data(
        'erika.frontend', 'data/episode_details_template.html'
    ).decode('utf-8')
    PODCAST_TEMPLATE = pkgutil.get_data(
        'erika.frontend', 'data/podcast_details_template.html'
    ).decode('utf-8')

    def __init__(self):
        super().__init__()

        self.current = None  # Current episode or podcast being shown

        self.view = WebView()
        self.add(self.view)

    def show_episode(self, episode, get_image=True):
        """Show the details of an episode"""
        self.current = episode

        html = self.EPISODE_TEMPLATE.format(episode=episode)
        self.view.load_html_string(html, "")

        # If the image is not available, download it in a thread, and
        # update the view at the end
        application = Gio.Application.get_default()
        if all((get_image,
                not episode.image_downloaded,
                application.get_online())):
            def _end():
                if episode == self.current:
                    self.show_episode(episode, get_image=False)

            def _thread():
                episode.get_image()
                GObject.idle_add(_end)

            thread = threading.Thread(target=_thread)
            thread.start()

    def show_podcast(self, podcast):
        """Show the details of a podcast"""
        self.current = podcast

        html = self.PODCAST_TEMPLATE.format(podcast=podcast)
        self.view.load_html_string(html, "")



# -*- coding: utf-8 -*-
#
# Copyright (c) 2016 Muges
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Module defining the database
"""

import string
from playhouse.sqlite_ext import SqliteExtDatabase
from peewee import Model

database = SqliteExtDatabase(None)  # pylint: disable=invalid-name


class BaseModel(Model):
    """
    Base model defining which database to use.
    """
    class Meta:  # pylint: disable=too-few-public-methods, missing-docstring
        database = database


@database.func()
def slugify(name):
    """Remove non alphanumeric characters from a string and convert it to
    lowercase"""
    return ''.join(
        char for char in name.lower()
        if char in string.ascii_lowercase + string.digits
    )



# -*- coding: utf-8 -*-
#
# Copyright (c) 2016 Muges
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Table containing the episode actions
"""

from peewee import TextField

from .database import BaseModel


class PodcastAction(BaseModel):
    """A model used to store podcast actions in the library.

    The podcast actions are used for the gpodder synchronization.

    Attributes
    ----------
    podcast_url : str
        The url of the podcast.
    action : str
        The type of the action (add or remove).
    """
    podcast_url = TextField(primary_key=True)
    action = TextField()

    @staticmethod
    def new(podcast_url, action):
        """Create a new podcast action (or overwrite an existing one)."""
        try:
            podcast_action = PodcastAction.get(podcast_url=podcast_url)
        except PodcastAction.DoesNotExist:
            podcast_action = PodcastAction.create(podcast_url=podcast_url, action=action)
        else:
            podcast_action.action = action
            podcast_action.save()


# -*- coding: utf-8 -*-
#
# Copyright (c) 2016 Muges
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
A Gtk.Label with sane defaults
"""

from gi.repository import Gtk
from gi.repository import Pango


class Label(Gtk.Label):
    """A Gtk.Label with sane defaults

    Creates a label aligned to the left, ellipsized at the end, and with
    optional line wrapping.

    Parameters
    ----------
    lines : Optional[int]
        If lines is set to an integer greater than 1, enable wrapping and limit
        the number of lines that should be displayed.
    """
    def __init__(self, lines=1):
        super().__init__()

        self.set_alignment(xalign=0, yalign=0.5)
        self.set_ellipsize(Pango.EllipsizeMode.END)

        if lines > 1:
            self.set_line_wrap(True)
            self.set_line_wrap_mode(Pango.WrapMode.WORD_CHAR)
            self.set_lines(lines)



# -*- coding: utf-8 -*-
#
# Copyright (c) 2016 Muges
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
The :mod:`erika.util` provides various utility functions.
"""

__all__ = [
    'guess_extension', 'sanitize_filename', 'format_duration',
    'format_fulltext_duration', 'format_size', 'plaintext_to_html',
    'check_connection', 'html_to_plaintext'
]

from .files import guess_extension, sanitize_filename
from .format import format_duration, format_fulltext_duration, format_size
from .plaintext_to_html import plaintext_to_html
from .network import check_connection
from .html_to_plaintext import html_to_plaintext



# -*- coding: utf-8 -*-
#
# Copyright (c) 2016 Muges
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Dialog used to edit the configuration
"""

from gi.repository import Gtk

from erika.library.models import database, Config
from .util import get_builder


class PreferencesDialog(Gtk.Dialog):
    """Dialog used to edit the configuration"""

    ENTRIES = ['library.podcast_directory_template',
               'library.episode_file_template', 'gpodder.deviceid',
               'gpodder.devicename', 'gpodder.password', 'gpodder.username',
               'gpodder.hostname']
    SPINS = ['library.synchronize_interval', 'downloads.workers',
             'player.smart_mark_seconds']
    CHECKS = ['gpodder.synchronize']
    FILES = ['library.root']

    def __init__(self, parent=None):
        Gtk.Dialog.__init__(self, "Preferences", parent, 0,
                            (Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
                             Gtk.STOCK_APPLY, Gtk.ResponseType.OK))

        # Initialize dialog and widgets
        self.builder = get_builder('data/preferences.glade')
        content = self.builder.get_object("dialog-content")
        self.get_content_area().add(content)

    def save_config(self):
        """Save the configuration in the database"""
        with database.transaction():
            for key in self.ENTRIES:
                entry = self.builder.get_object(key)
                Config.set_value(key, entry.get_text())

            for key in self.SPINS:
                spinbutton = self.builder.get_object(key)
                Config.set_value(key, spinbutton.get_value_as_int())

            for key in self.CHECKS:
                checkbutton = self.builder.get_object(key)
                Config.set_value(key, checkbutton.get_active())

            for key in self.FILES:
                filechooser = self.builder.get_object(key)
                Config.set_value(key, filechooser.get_filename())

    def load_config(self):
        """Load the configuration from the database"""
        for key in self.ENTRIES:
            entry = self.builder.get_object(key)
            entry.set_text(Config.get_value(key))

        for key in self.SPINS:
            spinbutton = self.builder.get_object(key)
            spinbutton.set_value(Config.get_value(key))

        for key in self.CHECKS:
            checkbutton = self.builder.get_object(key)
            checkbutton.set_active(Config.get_value(key))

        for key in self.FILES:
            filechooser = self.builder.get_object(key)
            filechooser.set_filename(Config.get_value(key))


def run(window=None):
    """Show the dialog and edit the configuration"""
    dialog = PreferencesDialog(window)
    dialog.load_config()

    response = dialog.run()
    if response == Gtk.ResponseType.OK:
        dialog.save_config()

    dialog.destroy()



# -*- coding: utf-8 -*-
#
# Copyright (c) 2016 Muges
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
List of episodes
"""

from gi.repository import GObject
from gi.repository import GLib
from gi.repository import Gtk
from gi.repository import Gst
from gi.repository import Gio

from erika.library.models import database, Episode, EpisodeAction
from erika.util import format_duration
from .widgets import Label, IndexedListBox, FilterButton, SortButton
from .player import Player
from .util import cb

SUBTITLE_LINES = 4
CHUNK_SIZE = 10


class EpisodeList(Gtk.VBox):
    """List of episodes

    Signals
    -------
    episodes-changed
        Emitted when episodes are modified (e.g. marked as played)
    download(Episode)
        Emitted when the download button of an episode is clicked
    """

    __gsignals__ = {
        'episodes-changed':
            (GObject.SIGNAL_RUN_FIRST, None, ()),
        'download':
            (GObject.SIGNAL_RUN_FIRST, None, (GObject.TYPE_PYOBJECT,)),
        'episode-selected':
            (GObject.SIGNAL_RUN_FIRST, None, (GObject.TYPE_PYOBJECT,)),
        'podcast-selected':
            (GObject.SIGNAL_RUN_FIRST, None, (GObject.TYPE_PYOBJECT,)),
    }

    def __init__(self, player):
        Gtk.VBox.__init__(self)

        self.current_podcast = None
        self.update_id = None

        self.player = player

        # Episode list
        self.listbox = IndexedListBox()
        self.listbox.set_selection_mode(Gtk.SelectionMode.MULTIPLE)
        self.listbox.set_sort_func(self.sort_func)
        self.listbox.set_filter_func(self.filter_func)
        self.listbox.connect("popup-menu", self._on_popup_menu)
        self.listbox.connect("selected-rows-changed",
                             self._on_selected_rows_changed)

        scrolled_window = Gtk.ScrolledWindow()
        scrolled_window.add_with_viewport(self.listbox)

        # Filter and sort buttons
        action_bar = Gtk.ActionBar()

        filter_new = FilterButton("erika-episode-new-symbolic",
                                  "new episodes",
                                  lambda episode: episode.new)
        filter_new.connect("toggled", cb(self.listbox.invalidate_filter))

        filter_played = FilterButton("media-playback-start-symbolic",
                                     "played episodes",
                                     lambda episode: episode.played)
        filter_played.connect("toggled", cb(self.listbox.invalidate_filter))

        filter_downloaded = FilterButton("document-save-symbolic",
                                         "downloaded episodes",
                                         lambda episode: episode.downloaded)
        filter_downloaded.connect("toggled",
                                  cb(self.listbox.invalidate_filter))

        self.filters = [filter_new, filter_played, filter_downloaded]

        self.sort = SortButton("publication date")
        self.sort.connect("clicked", cb(self.listbox.invalidate_sort))

        # Layout
        action_bar.pack_start(filter_new)
        action_bar.pack_start(filter_played)
        action_bar.pack_start(filter_downloaded)
        action_bar.pack_end(self.sort)

        self.pack_start(action_bar, False, False, 0)
        self.pack_start(scrolled_window, True, True, 0)

        self.application = Gio.Application.get_default()
        self.application.connect('network-state-changed',
                                 cb(self.set_network_state))

    def select(self, podcast):
        """Display the episodes of a podcast

        Parameters
        ----------
        podcast : Podcast
            The podcast whose episodes will be displayed
        """
        self.current_podcast = podcast

        # Remove children
        self.listbox.clear()

        self.update()

    def update(self):
        """Update the list of episodes of the current podcast"""
        if self.current_podcast:
            # Interrupt the previous update
            if self.update_id:
                GLib.source_remove(self.update_id)

            episodes = self.current_podcast.episodes

            order_clauses = []

            # Put the results that are not shown at the end (so that
            # they can be shown if the user changes the filter, but do
            # not delay the display of the ones that are currently
            # visible)
            for filter_button in self.filters:
                if filter_button.state is True:
                    order_clauses.append(filter_button.key(Episode).desc())
                elif filter_button.state is False:
                    order_clauses.append(filter_button.key(Episode).asc())

            # Sort by date
            if self.sort.get_descending():
                order_clauses.append(Episode.pubdate.desc())
            else:
                order_clauses.append(Episode.pubdate.asc())

            episodes = episodes.order_by(*order_clauses)

            self._load_by_chunks(iter(episodes), self.listbox.get_ids())

    def update_episode(self, episode):
        """Update an episode

        Parameters
        ----------
        episode : Episode
        """
        try:
            row = self.listbox.get_row(episode.id)
        except ValueError:
            pass
        else:
            row.episode = episode
            row.update()

    def _load_by_chunks(self, episodes, remove_ids):
        """Load episodes by chunks of size CHUNK_SIZE to prevent rendering
        delay

        Parameters
        ----------
        episodes : Iterable[Episode]
            The episodes which will be displayed
        remove_ids : Set
            Set of ids that were in the list but have not been updated yet
            (these should be removed of the list at the end)

        """
        for _ in range(0, CHUNK_SIZE):
            try:
                episode = next(episodes)
            except StopIteration:
                # Remove the episodes that are not in the database anymore
                for episode_id in remove_ids:
                    self.listbox.remove_id(episode_id)

                self.update_id = None
                return
            else:
                try:
                    row = self.listbox.get_row(episode.id)
                except ValueError:
                    # The row does not exist, create it
                    row = EpisodeRow(episode)
                    row.connect("toggle", self._toggle)
                    row.connect("download", self._download)
                    row.set_online(self.application.get_online())

                    if episode == self.player.episode:
                        # The episode is currently being played
                        row.set_progress(self.player.get_seconds_position(),
                                         self.player.get_seconds_duration())
                        row.set_player_state(self.player.state)

                    self.listbox.add_with_id(row, episode.id)
                else:
                    # The row already exists, update it
                    remove_ids.remove(episode.id)
                    row.episode = episode
                    row.update()

        # Load the rest of the episodes in a future iteration of the main
        # loop (after this chunk has been rendered).
        self.update_id = GLib.idle_add(self._load_by_chunks, episodes,
                                       remove_ids)

    def sort_func(self, row1, row2):
        """Compare two rows to determine which should be first. Sort them
        by date

        Parameters
        ----------
        row1 : Gtk.ListBoxRow
            The first row
        row2 : Gtk.ListBoxRow
            The second row

        Returns
        -------
        int
            -1 if row1 should be before row2,
            0 if they are equal,
            1 otherwise
        """
        if self.sort.get_descending():
            delta = 1
        else:
            delta = -1

        if row1.episode.pubdate < row2.episode.pubdate:
            return delta
        elif row1.episode.pubdate == row2.episode.pubdate:
            return 0
        else:
            return -delta

    def filter_func(self, row):
        """Check if a row should be visible or not

        Parameters
        ----------
        row : Gtk.ListBoxRow
            The row

        Returns
        -------
        bool
            True if the row should be visible.
        """
        for filter_button in self.filters:
            state = filter_button.state
            if state is not None and filter_button.key(row.episode) != state:
                return False

        return True

    def _on_popup_menu(self, listbox):
        """Display the context menu"""
        selection = self.listbox.get_selected_rows()

        if selection == []:
            return

        menu = Gtk.Menu()

        if any(not row.episode.local_path for row in selection):
            menu_item = Gtk.MenuItem("Download")
            menu_item.connect("activate", cb(self._download_selection),
                              selection)
            menu.append(menu_item)

            menu_item = Gtk.MenuItem("Import file")
            menu_item.connect("activate", cb(self._import_file),
                              selection)
            menu.append(menu_item)

        if any(not row.episode.played for row in selection):
            menu_item = Gtk.MenuItem("Mark as played")
            menu_item.connect("activate", cb(self._mark_as_played),
                              selection)
            menu.append(menu_item)

        if any(row.episode.played for row in selection):
            menu_item = Gtk.MenuItem("Mark as unplayed")
            menu_item.connect("activate", cb(self._mark_as_unplayed),
                              selection)
            menu.append(menu_item)

        if any(row.episode.progress > 0 for row in selection):
            menu_item = Gtk.MenuItem("Reset progress")
            menu_item.connect("activate", cb(self._reset_progress),
                              selection)
            menu.append(menu_item)

        menu.show_all()
        menu.popup(None, None, None, None, 0, Gtk.get_current_event_time())

    def _mark_as_played(self, rows):
        """Mark selected rows as unplayed

        Parameters
        ----------
        rows : List[EpisodeRow]
            The selected rows
        """
        with database.transaction():
            for row in rows:
                row.episode.mark_as_played()
                row.update()
                row.episode.save()

                action = EpisodeAction(episode=row.episode, action="play",
                                       started=row.episode.duration,
                                       position=row.episode.duration,
                                       total=row.episode.duration)
                action.save()

        self.emit("episodes-changed")

    def _mark_as_unplayed(self, rows):
        """Mark selected rows as unplayed

        Parameters
        ----------
        rows : List[EpisodeRow]
            The selected rows
        """
        with database.transaction():
            for row in rows:
                row.episode.mark_as_unplayed()
                row.update()
                row.episode.save()

                action = EpisodeAction(episode=row.episode, action="new")
                action.save()

        self.emit("episodes-changed")

    def _reset_progress(self, rows):
        """Reset the progresses of the selected rows

        Parameters
        ----------
        rows : List[EpisodeRow]
            The selected rows
        """
        with database.transaction():
            for row in rows:
                row.episode.progress = 0
                row.update()
                row.episode.save()

                action = EpisodeAction(episode=row.episode, action="play",
                                       started=0, position=0,
                                       total=row.episode.duration)
                action.save()

        self.emit("episodes-changed")

    def _toggle(self, row):
        """Called when the toggle button of an episode is clicked"""
        if self.player.episode == row.episode:
            self.player.toggle()
        else:
            self.player.play_episode(row.episode)

    def _download(self, row):
        """Called when the download button of an episode is clicked"""
        self.emit("download", row.episode)

    def _download_selection(self, rows):
        """Download the episodes of the selected rows

        Parameters
        ----------
        rows : List[EpisodeRow]
            The selected rows
        """
        for row in rows:
            if row.episode.local_path is None:
                self.emit("download", row.episode)

    def _import_file(self, rows):
        """Import files for the episodes of the selected rows

        Parameters
        ----------
        rows : List[EpisodeRow]
            The selected rows
        """
        # Get parent window
        window = self.get_toplevel()

        for row in rows:
            if row.episode.local_path is None:
                dialog = Gtk.FileChooserDialog(
                    '"{}" file selection'.format(row.episode.title),
                    window, Gtk.FileChooserAction.OPEN,
                    (Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
                     Gtk.STOCK_OPEN, Gtk.ResponseType.OK))
                dialog.set_current_folder(row.episode.podcast.get_directory())

                response = dialog.run()
                if response == Gtk.ResponseType.OK:
                    path = dialog.get_filename()
                elif response == Gtk.ResponseType.CANCEL:
                    path = None
                dialog.destroy()

                if path:
                    row.episode.absolute_local_path = row.episode.import_file(path)
                    row.episode.save()
                    row.update()

                    action = EpisodeAction(episode=row.episode,
                                           action="download")
                    action.save()

    def set_player_progress(self, position, duration):
        """Called when the progress of the playback changes. Updates the
        row of the episode being played"""
        try:
            row = self.listbox.get_row(self.player.episode.id)
        except ValueError:
            pass
        else:
            row.set_progress(position // Gst.SECOND, duration // Gst.SECOND)

    def set_player_state(self, episode, state):
        """Called when the state of the player changes. Updates the
        row of the episode being played"""
        try:
            row = self.listbox.get_row(episode.id)
        except ValueError:
            pass
        else:
            row.set_player_state(state)

    def set_network_state(self, online, _):
        """Called when the network state changes"""
        for row in self.listbox.rows.values():
            row.set_online(online)

    def _on_selected_rows_changed(self, listbox):
        rows = self.listbox.get_selected_rows()
        if len(rows) == 1:
            self.emit("episode-selected", rows[0].episode)
        else:
            self.emit("podcast-selected", self.current_podcast)


class EpisodeRow(Gtk.ListBoxRow):
    # pylint: disable=too-many-instance-attributes
    """Row in the list of episodes

    Signals
    -------
    toggle
        Emitted when the toggle button is clicked
    download
        Emitted when the download button is clicked
    """

    __gsignals__ = {
        'toggle': (GObject.SIGNAL_RUN_FIRST, None, ()),
        'download': (GObject.SIGNAL_RUN_FIRST, None, ()),
    }

    def __init__(self, episode):
        Gtk.ListBoxRow.__init__(self)

        self.episode = episode

        grid = Gtk.Grid()
        grid.set_column_spacing(5)
        self.add(grid)

        # Episode status
        self.icon = Gtk.Image()
        self.icon.set_margin_start(7)
        self.icon.set_margin_end(7)
        grid.attach(self.icon, 0, 0, 1, 1)

        topgrid = Gtk.Grid()
        grid.attach(topgrid, 1, 0, 1, 1)

        bottomgrid = Gtk.Grid()
        grid.attach(bottomgrid, 1, 1, 1, 1)

        # Episode title
        self.title = Label()
        self.title.set_hexpand(True)
        self.title.set_margin_top(3)
        topgrid.attach(self.title, 0, 0, 1, 1)

        # Publication date
        self.date = Label()
        self.date.set_hexpand(True)
        topgrid.attach(self.date, 0, 1, 1, 1)

        # Download button
        self.download_button = Gtk.Button.new_from_icon_name(
            "document-save-symbolic", Gtk.IconSize.BUTTON)
        self.download_button.set_tooltip_text("Download the episode")
        self.download_button.set_relief(Gtk.ReliefStyle.NONE)
        self.download_button.connect('clicked', cb(self.emit), "download")
        topgrid.attach(self.download_button, 1, 0, 1, 2)

        # Play button
        self.toggle_button = Gtk.Button()
        self.toggle_button.set_relief(Gtk.ReliefStyle.NONE)
        self.toggle_button.connect('clicked', cb(self.emit), "toggle")
        topgrid.attach(self.toggle_button, 2, 0, 1, 2)

        # Episode subtitle
        self.subtitle = Label(lines=SUBTITLE_LINES)
        self.subtitle.set_margin_top(7)
        self.subtitle.set_margin_bottom(7)
        bottomgrid.attach(self.subtitle, 0, 0, 2, 1)

        # Episode duration
        self.duration = Label()
        bottomgrid.attach(self.duration, 0, 1, 1, 1)

        # Playback progress
        self.progress = Gtk.ProgressBar()
        self.progress.set_hexpand(True)
        self.progress.set_valign(Gtk.Align.CENTER)
        self.progress.set_margin_start(20)
        self.progress.set_margin_end(20)
        bottomgrid.attach(self.progress, 1, 1, 1, 1)

        # Separator
        separator = Gtk.Separator.new(Gtk.Orientation.HORIZONTAL)
        separator.set_margin_top(5)
        grid.attach(separator, 0, 2, 2, 1)

        self.show_all()

        self.update()

    def update(self):
        """Update the widget"""
        # Episode status
        if self.episode.new:
            self.icon.set_from_icon_name("erika-episode-new",
                                         Gtk.IconSize.BUTTON)
            self.icon.set_tooltip_text("New episode")
        else:
            self.icon.set_from_icon_name("erika-episode-old",
                                         Gtk.IconSize.BUTTON)
            self.icon.set_tooltip_text("Old episode")

        # Episode title
        self.title.set_sensitive(not self.episode.played)
        self.title.set_markup("<b>{}</b>".format(
            GLib.markup_escape_text(self.episode.title)))

        # Publication date
        self.date.set_sensitive(not self.episode.played)
        self.date.set_text("{:%d/%m/%Y}".format(self.episode.pubdate))

        # Episode subtitle
        self.subtitle.set_sensitive(not self.episode.played)
        self.subtitle.set_text(self.episode.subtitle)

        # Episode duration
        self.duration.set_sensitive(not self.episode.played)

        self.set_progress(self.episode.progress, self.episode.duration)
        self.set_player_state(Player.STOPPED)

        # Download button
        self.download_button.set_visible(not self.episode.local_path)

    def set_progress(self, position, duration):
        """Set the progress of the playback of the episode"""
        # Episode duration
        if duration:
            self.duration.set_text(format_duration(duration))
            self.duration.show()
        else:
            self.duration.hide()

        # Playback progress
        if duration and position > 0:
            self.progress.show()
            self.progress.set_fraction(position / duration)
        else:
            self.progress.hide()

    def set_player_state(self, state):
        """Set the state of the player for this episode"""
        if state == Player.PLAYING or state == Player.BUFFERING:
            self.toggle_button.set_image(
                Gtk.Image.new_from_icon_name("media-playback-pause-symbolic",
                                             Gtk.IconSize.BUTTON))
            self.toggle_button.set_tooltip_text("Pause the episode")
        else:
            self.toggle_button.set_image(
                Gtk.Image.new_from_icon_name("media-playback-start-symbolic",
                                             Gtk.IconSize.BUTTON))
            self.toggle_button.set_tooltip_text("Play the episode")

    def set_online(self, online):
        """Called when the network state changes"""
        self.download_button.set_sensitive(online)
        self.toggle_button.set_sensitive(online or self.episode.downloaded)
